from enum import IntEnum

from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.script.array import Array
from sonolus.script.array_like import ArrayLike
from sonolus.script.containers import ArrayPointer
from sonolus.script.globals import (
    _level_life,
    _level_score,
    _play_runtime_environment,
    _play_runtime_ui,
    _play_runtime_ui_configuration,
    _play_runtime_update,
    _preview_runtime_canvas,
    _preview_runtime_environment,
    _preview_runtime_ui,
    _preview_runtime_ui_configuration,
    _runtime_background,
    _runtime_particle_transform,
    _runtime_skin_transform,
    _runtime_touch_array,
    _tutorial_instruction,
    _tutorial_runtime_environment,
    _tutorial_runtime_ui,
    _tutorial_runtime_ui_configuration,
    _tutorial_runtime_update,
    _watch_runtime_environment,
    _watch_runtime_ui,
    _watch_runtime_ui_configuration,
    _watch_runtime_update,
)
from sonolus.script.internal.context import ctx
from sonolus.script.internal.meta_fn import meta_fn, perf_meta_fn
from sonolus.script.internal.native import native_function
from sonolus.script.num import Num
from sonolus.script.quad import Quad, Rect
from sonolus.script.record import Record
from sonolus.script.transform import Transform2d
from sonolus.script.vec import Vec2


@_play_runtime_environment
class _PlayRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float
    input_offset: float
    is_multiplayer: bool


@_watch_runtime_environment
class _WatchRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float
    input_offset: float
    is_replay: bool


@_preview_runtime_environment
class _PreviewRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float


@_tutorial_runtime_environment
class _TutorialRuntimeEnvironment:
    is_debug: bool
    aspect_ratio: float
    audio_offset: float


@_play_runtime_update
class _PlayRuntimeUpdate:
    time: float
    delta_time: float
    scaled_time: float
    touch_count: int


@_watch_runtime_update
class _WatchRuntimeUpdate:
    time: float
    delta_time: float
    scaled_time: float
    is_skip: bool


@_tutorial_runtime_update
class _TutorialRuntimeUpdate:
    time: float
    delta_time: float
    navigation_direction: int


class ScrollDirection(IntEnum):
    LEFT_TO_RIGHT = 0
    TOP_TO_BOTTOM = 1
    RIGHT_TO_LEFT = 2
    BOTTOM_TO_TOP = 3


@_preview_runtime_canvas
class _PreviewRuntimeCanvas:
    scroll_direction: ScrollDirection
    size: float

    def update(self, scroll_direction: ScrollDirection | None = None, size: float | None = None):
        if scroll_direction is not None:
            self.scroll_direction = scroll_direction
        if size is not None:
            self.size = size


class RuntimeUiConfig(Record):
    scale: float
    alpha: float


@_play_runtime_ui_configuration
class _PlayRuntimeUiConfigs:
    menu: RuntimeUiConfig
    judgment: RuntimeUiConfig
    combo: RuntimeUiConfig
    primary_metric: RuntimeUiConfig
    secondary_metric: RuntimeUiConfig


@_watch_runtime_ui_configuration
class _WatchRuntimeUiConfigs:
    menu: RuntimeUiConfig
    judgment: RuntimeUiConfig
    combo: RuntimeUiConfig
    primary_metric: RuntimeUiConfig
    secondary_metric: RuntimeUiConfig
    progress: RuntimeUiConfig
    progress_graph: RuntimeUiConfig


@_preview_runtime_ui_configuration
class _PreviewRuntimeUiConfigs:
    menu: RuntimeUiConfig
    progress: RuntimeUiConfig


@_tutorial_runtime_ui_configuration
class _TutorialRuntimeUiConfigs:
    menu: RuntimeUiConfig
    navigation: RuntimeUiConfig
    instruction: RuntimeUiConfig


class HorizontalAlign(IntEnum):
    LEFT = -1
    CENTER = 0
    RIGHT = 1


class RuntimeUiLayout(Record):
    anchor: Vec2
    pivot: Vec2
    dimensions: Vec2
    rotation: float
    alpha: float
    horizontal_align: HorizontalAlign
    background: bool

    def update(
        self,
        anchor: Vec2 | None = None,
        pivot: Vec2 | None = None,
        dimensions: Vec2 | None = None,
        rotation: float | None = None,
        alpha: float | None = None,
        horizontal_align: HorizontalAlign | None = None,
        background: bool | None = None,
    ):
        if anchor is not None:
            self.anchor = anchor
        if pivot is not None:
            self.pivot = pivot
        if dimensions is not None:
            self.dimensions = dimensions
        if rotation is not None:
            self.rotation = rotation
        if alpha is not None:
            self.alpha = alpha
        if horizontal_align is not None:
            self.horizontal_align = horizontal_align
        if background is not None:
            self.background = background


class BasicRuntimeUiLayout(Record):
    anchor: Vec2
    pivot: Vec2
    dimensions: Vec2
    rotation: float
    alpha: float
    background: bool

    def update(
        self,
        anchor: Vec2 | None = None,
        pivot: Vec2 | None = None,
        dimensions: Vec2 | None = None,
        rotation: float | None = None,
        alpha: float | None = None,
        background: bool | None = None,
    ):
        if anchor is not None:
            self.anchor = anchor
        if pivot is not None:
            self.pivot = pivot
        if dimensions is not None:
            self.dimensions = dimensions
        if rotation is not None:
            self.rotation = rotation
        if alpha is not None:
            self.alpha = alpha
        if background is not None:
            self.background = background


@_play_runtime_ui
class _PlayRuntimeUi:
    menu: RuntimeUiLayout
    judgment: RuntimeUiLayout
    combo_value: RuntimeUiLayout
    combo_text: RuntimeUiLayout
    primary_metric_bar: RuntimeUiLayout
    primary_metric_value: RuntimeUiLayout
    secondary_metric_bar: RuntimeUiLayout
    secondary_metric_value: RuntimeUiLayout


@_watch_runtime_ui
class _WatchRuntimeUi:
    menu: RuntimeUiLayout
    judgment: RuntimeUiLayout
    combo_value: RuntimeUiLayout
    combo_text: RuntimeUiLayout
    primary_metric_bar: RuntimeUiLayout
    primary_metric_value: RuntimeUiLayout
    secondary_metric_bar: RuntimeUiLayout
    secondary_metric_value: RuntimeUiLayout
    progress: RuntimeUiLayout
    progress_graph: RuntimeUiLayout


@_preview_runtime_ui
class _PreviewRuntimeUi:
    menu: BasicRuntimeUiLayout
    progress: BasicRuntimeUiLayout


@_tutorial_runtime_ui
class _TutorialRuntimeUi:
    menu: BasicRuntimeUiLayout
    previous: BasicRuntimeUiLayout
    next: BasicRuntimeUiLayout
    instruction: BasicRuntimeUiLayout


class UiLayout[T](Record):
    """The layout of a UI element."""

    _underlying: T

    def update(
        self,
        anchor: Vec2 | None = None,
        pivot: Vec2 | None = None,
        dimensions: Vec2 | None = None,
        rotation: float | None = None,
        alpha: float | None = None,
        horizontal_align: HorizontalAlign | None = None,
        background: bool | None = None,
    ):
        """Update the layout properties if it's available in the current mode and do nothing otherwise."""
        match self._underlying:
            case RuntimeUiLayout():
                self._underlying.update(
                    anchor=anchor,
                    pivot=pivot,
                    dimensions=dimensions,
                    rotation=rotation,
                    alpha=alpha,
                    horizontal_align=horizontal_align,
                    background=background,
                )
            case BasicRuntimeUiLayout():
                self._underlying.update(
                    anchor=anchor,
                    pivot=pivot,
                    dimensions=dimensions,
                    rotation=rotation,
                    alpha=alpha,
                    background=background,
                )
            case _:
                pass  # do nothing

    @property
    def is_available(self) -> bool:
        """Check if the layout is available in the current mode."""
        return self._underlying is not None


class UiConfig[T](Record):
    """The user configuration for a UI element."""

    _underlying: T

    @property
    def scale(self) -> float:
        """The scale of the UI element."""
        match self._underlying:
            case RuntimeUiConfig():
                return self._underlying.scale
            case _:
                return 1.0  # Default scale if not available

    @property
    def alpha(self) -> float:
        """The alpha (opacity) of the UI element."""
        match self._underlying:
            case RuntimeUiConfig():
                return self._underlying.alpha
            case _:
                return 1.0  # Default alpha if not available

    @property
    def is_available(self) -> bool:
        """Check if the config is available in the current mode."""
        return self._underlying is not None


class RuntimeUi(Record):
    """Holds the layouts for different UI elements across all modes."""

    @property
    @meta_fn
    def menu(self) -> UiLayout:
        """The configuration for the menu UI element.

        Available in play, watch, preview, and tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.menu)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.menu)
            case Mode.PREVIEW:
                return UiLayout(_PreviewRuntimeUi.menu)
            case Mode.TUTORIAL:
                return UiLayout(_TutorialRuntimeUi.menu)
            case _:
                raise RuntimeError("Unsupported mode for menu UI layout")

    @property
    @meta_fn
    def menu_config(self) -> UiConfig:
        """The configuration for the menu UI element.

        Available in play, watch, preview, and tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiConfig(_PlayRuntimeUiConfigs.menu)
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.menu)
            case Mode.PREVIEW:
                return UiConfig(_PreviewRuntimeUiConfigs.menu)
            case Mode.TUTORIAL:
                return UiConfig(_TutorialRuntimeUiConfigs.menu)
            case _:
                raise RuntimeError("Unsupported mode for menu UI configuration")

    @property
    @meta_fn
    def judgment(self) -> UiLayout:
        """The configuration for the judgment UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.judgment)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.judgment)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def judgment_config(self) -> UiConfig:
        """The configuration for the judgment UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiConfig(_PlayRuntimeUiConfigs.judgment)
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.judgment)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def combo_value(self) -> UiLayout:
        """The configuration for the combo value UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.combo_value)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.combo_value)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def combo_text(self) -> UiLayout:
        """The configuration for the combo text UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.combo_text)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.combo_text)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def combo_config(self) -> UiConfig:
        """The configuration for the combo UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiConfig(_PlayRuntimeUiConfigs.combo)
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.combo)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def primary_metric_bar(self) -> UiLayout:
        """The configuration for the primary metric bar UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.primary_metric_bar)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.primary_metric_bar)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def primary_metric_value(self) -> UiLayout:
        """The configuration for the primary metric value UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.primary_metric_value)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.primary_metric_value)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def primary_metric_config(self) -> UiConfig:
        """The configuration for the primary metric UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiConfig(_PlayRuntimeUiConfigs.primary_metric)
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.primary_metric)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def secondary_metric_bar(self) -> UiLayout:
        """The configuration for the secondary metric bar UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.secondary_metric_bar)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.secondary_metric_bar)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def secondary_metric_value(self) -> UiLayout:
        """The configuration for the secondary metric value UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiLayout(_PlayRuntimeUi.secondary_metric_value)
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.secondary_metric_value)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def secondary_metric_config(self) -> UiConfig:
        """The configuration for the secondary metric UI element.

        Available in play and watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.PLAY:
                return UiConfig(_PlayRuntimeUiConfigs.secondary_metric)
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.secondary_metric)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def progress(self) -> UiLayout:
        """The configuration for the progress UI element.

        Available in watch and preview mode.
        """
        match ctx().mode_state.mode:
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.progress)
            case Mode.PREVIEW:
                return UiLayout(_PreviewRuntimeUi.progress)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def progress_graph(self) -> UiLayout:
        """The configuration for the progress graph UI element.

        Available in watch mode.
        """
        match ctx().mode_state.mode:
            case Mode.WATCH:
                return UiLayout(_WatchRuntimeUi.progress_graph)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def progress_config(self) -> UiConfig:
        """The configuration for the progress UI element.

        Available in watch and preview mode.
        """
        match ctx().mode_state.mode:
            case Mode.WATCH:
                return UiConfig(_WatchRuntimeUiConfigs.progress)
            case Mode.PREVIEW:
                return UiConfig(_PreviewRuntimeUiConfigs.progress)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def previous(self) -> UiLayout:
        """The configuration for the previous navigation UI element.

        Available in tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.TUTORIAL:
                return UiLayout(_TutorialRuntimeUi.previous)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def next(self) -> UiLayout:
        """The configuration for the next navigation UI element.

        Available in tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.TUTORIAL:
                return UiLayout(_TutorialRuntimeUi.next)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def navigation_config(self) -> UiConfig:
        """The configuration for the navigation UI element.

        Available in tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.TUTORIAL:
                return UiConfig(_TutorialRuntimeUiConfigs.navigation)
            case _:
                return UiConfig(None)

    @property
    @meta_fn
    def instruction(self) -> UiLayout:
        """The configuration for the instruction UI element.

        Available in tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.TUTORIAL:
                return UiLayout(_TutorialRuntimeUi.instruction)
            case _:
                return UiLayout(None)

    @property
    @meta_fn
    def instruction_config(self) -> UiConfig:
        """The configuration for the instruction UI element.

        Available in tutorial mode.
        """
        match ctx().mode_state.mode:
            case Mode.TUTORIAL:
                return UiConfig(_TutorialRuntimeUiConfigs.instruction)
            case _:
                return UiConfig(None)


class Touch(Record):
    """Data of a touch event."""

    id: int
    """The unique identifier of the touch."""

    started: bool
    """Whether the touch has started this frame."""

    ended: bool
    """Whether the touch has ended this frame."""

    time: float
    """The time of the touch event.

    May remain constant while there is no movement.
    """

    start_time: float
    """The time the touch started."""

    position: Vec2
    """The current position of the touch."""

    start_position: Vec2
    """The position the touch started."""

    delta: Vec2
    """The change in position of the touch."""

    velocity: Vec2
    """The velocity of the touch."""

    speed: float
    """The speed of the touch's movement."""

    angle: float
    """The angle of the touch's movement."""

    @property
    def prev_position(self) -> Vec2:
        """The previous position of the touch."""
        return self.position - self.delta

    @property
    def total_delta(self) -> Vec2:
        """The total change in position of the touch."""
        return self.position - self.start_position

    @property
    def total_angle(self) -> float:
        """The total angle of the touch's movement."""
        return self.total_delta.angle


@_runtime_touch_array
class _TouchArray:
    # Handled specially, see touches()
    pass


@_runtime_skin_transform
class _SkinTransform:
    value: Array[Array[float, 4], 4]  # type: ignore

    @property
    @meta_fn
    def transform(self) -> Transform2d:
        values = self.value._to_list_()
        return Transform2d._raw(
            a00=Num(values[0 * 4 + 0]),
            a01=Num(values[0 * 4 + 1]),
            a02=Num(values[0 * 4 + 3]),
            a10=Num(values[1 * 4 + 0]),
            a11=Num(values[1 * 4 + 1]),
            a12=Num(values[1 * 4 + 3]),
            a20=Num(values[3 * 4 + 0]),
            a21=Num(values[3 * 4 + 1]),
            a22=Num(values[3 * 4 + 3]),
        )


@_runtime_particle_transform
class _ParticleTransform:
    value: Array[Array[float, 4], 4]  # type: ignore

    @property
    @meta_fn
    def transform(self) -> Transform2d:
        values = self.value._to_list_()
        return Transform2d._raw(
            a00=Num(values[0 * 4 + 0]),
            a01=Num(values[0 * 4 + 1]),
            a02=Num(values[0 * 4 + 3]),
            a10=Num(values[1 * 4 + 0]),
            a11=Num(values[1 * 4 + 1]),
            a12=Num(values[1 * 4 + 3]),
            a20=Num(values[3 * 4 + 0]),
            a21=Num(values[3 * 4 + 1]),
            a22=Num(values[3 * 4 + 3]),
        )


@_runtime_background
class _Background:
    value: Quad


@_level_score
class _LevelScore:
    perfect_multiplier: float
    great_multiplier: float
    good_multiplier: float
    consecutive_perfect_multiplier: float
    consecutive_perfect_step: float
    consecutive_perfect_cap: float
    consecutive_great_multiplier: float
    consecutive_great_step: float
    consecutive_great_cap: float
    consecutive_good_multiplier: float
    consecutive_good_step: float
    consecutive_good_cap: float

    def update(
        self,
        perfect_multiplier: float | None = None,
        great_multiplier: float | None = None,
        good_multiplier: float | None = None,
        consecutive_perfect_multiplier: float | None = None,
        consecutive_perfect_step: float | None = None,
        consecutive_perfect_cap: float | None = None,
        consecutive_great_multiplier: float | None = None,
        consecutive_great_step: float | None = None,
        consecutive_great_cap: float | None = None,
        consecutive_good_multiplier: float | None = None,
        consecutive_good_step: float | None = None,
        consecutive_good_cap: float | None = None,
    ):
        if perfect_multiplier is not None:
            self.perfect_multiplier = perfect_multiplier
        if great_multiplier is not None:
            self.great_multiplier = great_multiplier
        if good_multiplier is not None:
            self.good_multiplier = good_multiplier
        if consecutive_perfect_multiplier is not None:
            self.consecutive_perfect_multiplier = consecutive_perfect_multiplier
        if consecutive_perfect_step is not None:
            self.consecutive_perfect_step = consecutive_perfect_step
        if consecutive_perfect_cap is not None:
            self.consecutive_perfect_cap = consecutive_perfect_cap
        if consecutive_great_multiplier is not None:
            self.consecutive_great_multiplier = consecutive_great_multiplier
        if consecutive_great_step is not None:
            self.consecutive_great_step = consecutive_great_step
        if consecutive_great_cap is not None:
            self.consecutive_great_cap = consecutive_great_cap
        if consecutive_good_multiplier is not None:
            self.consecutive_good_multiplier = consecutive_good_multiplier
        if consecutive_good_step is not None:
            self.consecutive_good_step = consecutive_good_step
        if consecutive_good_cap is not None:
            self.consecutive_good_cap = consecutive_good_cap


@_level_life
class _LevelLife:
    consecutive_perfect_increment: float
    consecutive_perfect_step: float
    consecutive_great_increment: float
    consecutive_great_step: float
    consecutive_good_increment: float
    consecutive_good_step: float
    initial: int
    maximum: int

    def update(
        self,
        consecutive_perfect_increment: float | None = None,
        consecutive_perfect_step: float | None = None,
        consecutive_great_increment: float | None = None,
        consecutive_great_step: float | None = None,
        consecutive_good_increment: float | None = None,
        consecutive_good_step: float | None = None,
        initial: int | None = None,
        maximum: int | None = None,
    ):
        if consecutive_perfect_increment is not None:
            self.consecutive_perfect_increment = consecutive_perfect_increment
        if consecutive_perfect_step is not None:
            self.consecutive_perfect_step = consecutive_perfect_step
        if consecutive_great_increment is not None:
            self.consecutive_great_increment = consecutive_great_increment
        if consecutive_great_step is not None:
            self.consecutive_great_step = consecutive_great_step
        if consecutive_good_increment is not None:
            self.consecutive_good_increment = consecutive_good_increment
        if consecutive_good_step is not None:
            self.consecutive_good_step = consecutive_good_step
        if initial is not None:
            self.initial = initial
        if maximum is not None:
            self.maximum = maximum


@_tutorial_instruction
class _TutorialInstruction:
    text_id: int


@meta_fn
def is_debug() -> bool:
    """Check if the game is running in debug mode."""
    if not ctx():
        return False
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeEnvironment.is_debug
        case Mode.WATCH:
            return _WatchRuntimeEnvironment.is_debug
        case Mode.PREVIEW:
            return _PreviewRuntimeEnvironment.is_debug
        case Mode.TUTORIAL:
            return _TutorialRuntimeEnvironment.is_debug
        case _:
            return False


@meta_fn
def is_play() -> bool:
    """Check if the game is running in play mode."""
    return bool(ctx() and ctx().mode_state.mode == Mode.PLAY)


@meta_fn
def is_preview() -> bool:
    """Check if the game is running in preview mode."""
    return bool(ctx() and ctx().mode_state.mode == Mode.PREVIEW)


@meta_fn
def is_watch() -> bool:
    """Check if the game is running in watch mode."""
    return bool(ctx() and ctx().mode_state.mode == Mode.WATCH)


@meta_fn
def is_tutorial() -> bool:
    """Check if the game is running in tutorial mode."""
    return bool(ctx() and ctx().mode_state.mode == Mode.TUTORIAL)


@meta_fn
def is_preprocessing() -> bool:
    """Check if the game is in the preprocessing stage.

    Returns True if the current callback is one of preprocess, spawn_order, spawn_time, or despawn_time.
    """
    return bool(ctx() and ctx().callback in {"preprocess", "spawnOrder", "spawnTime", "despawnTime"})


@meta_fn
def aspect_ratio() -> float:
    """Get the aspect ratio of the game."""
    if not ctx():
        return 16 / 9
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeEnvironment.aspect_ratio
        case Mode.WATCH:
            return _WatchRuntimeEnvironment.aspect_ratio
        case Mode.PREVIEW:
            return _PreviewRuntimeEnvironment.aspect_ratio
        case Mode.TUTORIAL:
            return _TutorialRuntimeEnvironment.aspect_ratio


@meta_fn
def audio_offset() -> float:
    """Get the audio offset of the game.

    Returns 0 in preview mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeEnvironment.audio_offset
        case Mode.WATCH:
            return _WatchRuntimeEnvironment.audio_offset
        case Mode.TUTORIAL:
            return _TutorialRuntimeEnvironment.audio_offset
        case _:
            return 0


@meta_fn
def input_offset() -> float:
    """Get the input offset of the game.

    Returns 0 in preview mode and tutorial mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeEnvironment.input_offset
        case Mode.WATCH:
            return _WatchRuntimeEnvironment.input_offset
        case _:
            return 0


@meta_fn
def is_multiplayer() -> bool:
    """Check if the game is running in multiplayer mode.

    Returns False if not in play mode.
    """
    if not ctx():
        return False
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeEnvironment.is_multiplayer
        case _:
            return False


@meta_fn
def is_replay() -> bool:
    """Check if the game is running in replay mode.

    Returns False if not in watch mode.
    """
    if not ctx():
        return False
    match ctx().mode_state.mode:
        case Mode.WATCH:
            return _WatchRuntimeEnvironment.is_replay
        case _:
            return False


@meta_fn
def time() -> float:
    """Get the current time of the game.

    Returns 0 in preview mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeUpdate.time
        case Mode.WATCH:
            return _WatchRuntimeUpdate.time
        case Mode.TUTORIAL:
            return _TutorialRuntimeUpdate.time
        case _:
            return 0


@meta_fn
def offset_adjusted_time() -> float:
    """Get the current time of the game adjusted by the input offset.

    Returns 0 in preview mode and the current time without adjustment in tutorial mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeUpdate.time - _PlayRuntimeEnvironment.input_offset
        case Mode.WATCH:
            return _WatchRuntimeUpdate.time - _WatchRuntimeEnvironment.input_offset
        case Mode.TUTORIAL:
            return _TutorialRuntimeUpdate.time
        case _:
            return 0


@meta_fn
def delta_time() -> float:
    """Get the time elapsed since the last frame.

    Returns 0 in preview mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeUpdate.delta_time
        case Mode.WATCH:
            return _WatchRuntimeUpdate.delta_time
        case Mode.TUTORIAL:
            return _TutorialRuntimeUpdate.delta_time
        case _:
            return 0


@meta_fn
def scaled_time() -> float:
    """Get the current time of the game affected by the time scale.

    Returns the unscaled time in tutorial mode and 0 in preview mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return _PlayRuntimeUpdate.scaled_time
        case Mode.WATCH:
            return _WatchRuntimeUpdate.scaled_time
        case Mode.TUTORIAL:
            return _TutorialRuntimeUpdate.time
        case _:
            return 0


@meta_fn
def prev_time() -> float:
    """Get the time of the previous frame.

    Returns 0 in preview mode.
    """
    return time() - delta_time()


@meta_fn
def touches() -> ArrayLike[Touch]:
    """Get the current touches of the game."""
    if not ctx():
        return Array[Touch, 0]()  # type: ignore
    match ctx().mode_state.mode:
        case Mode.PLAY:
            return ArrayPointer[Touch]._raw(
                size=Num._accept_(_PlayRuntimeUpdate.touch_count),
                block=Num._accept_(ctx().blocks.RuntimeTouchArray),
                offset=Num._accept_(0),
            )
        case _:
            return Array[Touch, 0]()  # type: ignore


@meta_fn
def is_skip() -> bool:
    """Check if there was a time skip this frame.

    Returns False if not in watch mode.
    """
    if not ctx():
        return False
    match ctx().mode_state.mode:
        case Mode.WATCH:
            return _WatchRuntimeUpdate.is_skip
        case _:
            return False


@meta_fn
def navigation_direction() -> int:
    """Get the navigation direction of the tutorial.

    Returns 0 if not in tutorial mode.
    """
    if not ctx():
        return 0
    match ctx().mode_state.mode:
        case Mode.TUTORIAL:
            return _TutorialRuntimeUpdate.navigation_direction
        case _:
            return 0


def skin_transform() -> Transform2d:
    """Get the global skin transform."""
    return _SkinTransform.transform  # type: ignore


@meta_fn
def set_skin_transform(value: Transform2d):
    """Set the global skin transform."""
    _SkinTransform.transform._copy_from_(value)  # type: ignore


def particle_transform() -> Transform2d:
    """Get the global particle transform."""
    return _ParticleTransform.transform  # type: ignore


@meta_fn
def set_particle_transform(value: Transform2d):
    """Set the global particle transform."""
    _ParticleTransform.transform._copy_from_(value)  # type: ignore


def background() -> Quad:
    """Get the background quad."""
    return _Background.value


def set_background(value: Quad):
    """Set the background quad."""
    _Background.value = value


play_ui = _PlayRuntimeUi
play_ui_configs = _PlayRuntimeUiConfigs
watch_ui = _WatchRuntimeUi
watch_ui_configs = _WatchRuntimeUiConfigs
preview_ui = _PreviewRuntimeUi
preview_ui_configs = _PreviewRuntimeUiConfigs
tutorial_ui = _TutorialRuntimeUi
tutorial_ui_configs = _TutorialRuntimeUiConfigs

_runtime_ui = RuntimeUi()


def runtime_ui() -> RuntimeUi:
    """Get the runtime UI configuration."""
    return _runtime_ui  # type: ignore


def canvas() -> _PreviewRuntimeCanvas:
    """Get the preview canvas."""
    return _PreviewRuntimeCanvas  # type: ignore


@perf_meta_fn
def screen() -> Rect:
    """Get the screen boundaries as a rectangle."""
    return Rect._unchecked(t=1, r=aspect_ratio(), b=-1, l=-aspect_ratio())


def level_score() -> _LevelScore:
    """Get the level score configuration."""
    return _LevelScore  # type: ignore


def level_life() -> _LevelLife:
    """Get the level life configuration."""
    return _LevelLife  # type: ignore


def add_life_scheduled(value: int, time: float):
    """Schedule adding life at a specific time.

    Args:
        value: The amount of life to add.
        time: The time to add the life at.
    """
    assert is_play() or is_watch(), "add_life can only be called in play or watch mode"
    _add_life_scheduled(value, time)


@native_function(Op.AddLifeScheduled)
def _add_life_scheduled(value: int, time: float):
    raise NotImplementedError



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import VarArray
from sonolus.script.internal.range import range_or_tuple
from tests.script.conftest import run_and_validate

ints = st.integers(min_value=-10, max_value=10)
floats = st.floats(min_value=-99999, max_value=99999, allow_nan=False, allow_infinity=False)


def test_tuple_destructure():
    def fn():
        t = (1, 2), (3, 4), 5
        (a, b), (c, d), e = t
        return Array(a, b, c, d, e)

    assert run_and_validate(fn) == Array(1, 2, 3, 4, 5)


def test_tuple_addition():
    def fn():
        t1 = 1, 2
        t2 = 3, 4
        (a, b), (c, d) = t1, t2
        return Array(a, b, c, d)

    assert run_and_validate(fn) == Array(1, 2, 3, 4)


@given(
    t1_list=st.lists(ints, min_size=0, max_size=10),
    t2_list=st.lists(ints, min_size=0, max_size=10),
)
def test_tuple_comparison(t1_list, t2_list):
    t1 = tuple(t1_list)
    t2 = tuple(t2_list)

    def fn():
        return Array(t1 == t2, t1 != t2, t1 < t2, t1 <= t2, t1 > t2, t1 >= t2)

    assert run_and_validate(fn) == Array(t1 == t2, t1 != t2, t1 < t2, t1 <= t2, t1 > t2, t1 >= t2)


@given(
    t_list=st.lists(ints, min_size=0, max_size=10),
)
def test_tuple_iteration(t_list):
    t = tuple(t_list)

    def fn():
        results = VarArray[int, len(t)].new()
        for v in t:
            results.append(v)
        return results

    assert list(run_and_validate(fn)) == list(t)


def test_heterogeneous_tuple_iteration():
    def fn():
        results = VarArray[int, 5].new()
        for v in ((1, 2), (3, 4), 5):
            if isinstance(v, tuple):
                for i in v:
                    results.append(i)
            else:
                results.append(v)
        return results

    assert list(run_and_validate(fn)) == [1, 2, 3, 4, 5]


def test_tuple_negative_indexing():
    def fn():
        t = (10, 20, 30, 40, 50)

        return Array(t[-1], t[-2], t[-3], t[-4], t[-5])

    assert tuple(run_and_validate(fn)) == (50, 40, 30, 20, 10)


def test_tuple_mixed_indexing():
    def fn():
        t = (100, 200, 300, 400, 500)

        return Array(t[0], t[-5], t[2], t[-3], t[4], t[-1])

    assert tuple(run_and_validate(fn)) == (100, 100, 300, 300, 500, 500)


@given(
    t_list=st.lists(ints, min_size=1, max_size=10),
)
def test_tuple_negative_positive_indexing_equivalence(t_list):
    t = tuple(t_list)
    n = len(t_list)

    def fn():
        results = VarArray[bool, n].new()
        for i in range_or_tuple(n):
            results.append(t[i] == t[i - n])

        return results

    assert all(run_and_validate(fn))


def test_nested_tuple_indexing():
    def fn():
        t = ((1, 2), (3, 4), (5, 6))

        return Array(t[0][0], t[-1][-1], t[1][-1], t[-2][0])

    assert tuple(run_and_validate(fn)) == (1, 6, 4, 3)


@given(
    values_list=st.lists(floats, min_size=1, max_size=10),
)
def test_max_tuples(values_list):
    values = tuple(values_list)

    def fn():
        return max(values)

    assert run_and_validate(fn) == max(values)


@given(
    values_list=st.lists(floats, min_size=1, max_size=10),
)
def test_min_tuples(values_list):
    values = tuple(values_list)

    def fn():
        return min(values)

    assert run_and_validate(fn) == min(values)


def test_tuple_contains_basic():
    def fn():
        t = (1, 2, 3, 4, 5)
        return Array(
            1 in t,
            5 in t,
            0 in t,
            6 in t,
            3 in t,
        )

    assert tuple(run_and_validate(fn)) == (True, True, False, False, True)


def test_tuple_contains_empty():
    def fn():
        t = ()
        return Array(
            1 in t,
            0 in t,
        )

    assert tuple(run_and_validate(fn)) == (False, False)


def test_tuple_contains_single_element():
    def fn():
        t = (42,)
        return Array(
            42 in t,
            0 in t,
            1 in t,
        )

    assert tuple(run_and_validate(fn)) == (True, False, False)


def test_tuple_contains_duplicates():
    def fn():
        t = (1, 2, 2, 3, 3, 3)
        return Array(
            1 in t,
            2 in t,
            3 in t,
            4 in t,
        )

    assert tuple(run_and_validate(fn)) == (True, True, True, False)


def test_tuple_contains_heterogeneous():
    def fn():
        t = ((1, 2), (3, 4), 5)
        inner1 = (1, 2)
        inner2 = (3, 4)
        inner3 = (1, 3)
        return Array(
            5 in t,
            0 in t,
            inner1 in t,
            inner2 in t,
            inner3 in t,
        )

    assert tuple(run_and_validate(fn)) == (True, False, True, True, False)



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.liveness import HasLiveness, LivenessAnalysis, get_live, get_live_phi_targets
from sonolus.backend.optimize.passes import CompilerPass, OptimizerConfig
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class UnreachableCodeElimination(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        original_blocks = [*traverse_cfg_preorder(entry)]
        worklist = {entry}
        visited = set()
        while worklist:
            block = worklist.pop()
            if block in visited:
                continue
            visited.add(block)
            match block.test:
                case IRConst(value=value):
                    block.test = IRConst(0)
                    taken_edge = next(
                        (edge for edge in block.outgoing if edge.cond == value),
                        None,
                    ) or next((edge for edge in block.outgoing if edge.cond is None), None)
                    assert not block.outgoing or taken_edge
                    for edge in [*block.outgoing]:
                        if edge is not taken_edge:
                            edge.dst.incoming.remove(edge)
                            block.outgoing.remove(edge)
                    if taken_edge:
                        taken_edge.cond = None
                        worklist.add(taken_edge.dst)
                case _:
                    worklist.update(edge.dst for edge in block.outgoing)
        for block in original_blocks:
            if block not in visited:
                for edge in block.outgoing:
                    edge.dst.incoming.remove(edge)
            else:
                incoming_blocks = {edge.src for edge in block.incoming}
                for args in block.phis.values():
                    for src_block in [*args]:
                        if src_block not in visited or src_block not in incoming_blocks:
                            args.pop(src_block)
                block.phis = {tgt: args for tgt, args in block.phis.items() if args}
        return entry


class DeadCodeElimination(CompilerPass):
    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        uses = set()
        defs = {}
        for block in traverse_cfg_preorder(entry):
            for statement in block.statements:
                self.handle_statement(statement, uses, defs)
            for target, args in block.phis.items():
                if target not in defs:
                    defs[target] = []
                defs[target].append(tuple(args.values()))
            self.update_uses(block.test, uses)

        queue = [*uses]
        while queue:
            val = queue.pop()
            if val not in defs:
                continue
            for stmt in defs[val]:
                if isinstance(stmt, tuple):
                    stmt_uses = stmt
                else:
                    stmt_uses = self.update_uses(stmt, set())
                for use in stmt_uses:
                    if use not in uses:
                        uses.add(use)
                        queue.append(use)

        for block in traverse_cfg_preorder(entry):
            live_stmts = []
            for statement in block.statements:
                match statement:
                    case IRSet(place=place, value=value):
                        is_live = not (
                            (isinstance(place, SSAPlace) and place not in uses)
                            or (
                                isinstance(place, BlockPlace)
                                and isinstance(place.block, TempBlock)
                                and place.block not in uses
                            )
                            or (isinstance(value, IRGet) and place == value.place)
                        )
                        if is_live:
                            live_stmts.append(statement)
                        elif isinstance(value, IRInstr) and value.op.side_effects:
                            live_stmts.append(value)
                    case other:
                        live_stmts.append(other)
            block.statements = live_stmts
            block.phis = {place: phi for place, phi in block.phis.items() if place in uses}
        return entry

    def handle_statement(
        self,
        stmt: IRStmt | BlockPlace | SSAPlace | TempBlock | int,
        uses: set[HasLiveness],
        defs: dict[HasLiveness, list[IRStmt | tuple[HasLiveness]]],
    ):
        if isinstance(stmt, IRSet):
            place = stmt.place
            value = stmt.value
            if isinstance(place, SSAPlace):
                if place not in defs:
                    defs[place] = []
                defs[place].append(stmt)
                if isinstance(value, IRInstr) and value.op.side_effects:
                    self.update_uses(value, uses)
            elif isinstance(place, BlockPlace) and isinstance(place.block, TempBlock):
                if place.block not in defs:
                    defs[place.block] = []
                defs[place.block].append(stmt)
                if isinstance(value, IRInstr) and value.op.side_effects:
                    self.update_uses(value, uses)
            else:
                self.update_uses(place, uses)
                self.update_uses(value, uses)
        else:
            self.update_uses(stmt, uses)

    def update_uses(
        self, stmt: IRStmt | BlockPlace | SSAPlace | TempBlock | int, uses: set[HasLiveness]
    ) -> set[HasLiveness]:
        match stmt:
            case IRPureInstr(op=_, args=args) | IRInstr(op=_, args=args):
                for arg in args:
                    self.update_uses(arg, uses)
            case IRGet(place=place):
                self.update_uses(place, uses)
            case IRSet(place=place, value=value):
                if isinstance(place, BlockPlace):
                    if not isinstance(place.block, TempBlock):
                        self.update_uses(place.block, uses)
                    self.update_uses(place.index, uses)
                self.update_uses(value, uses)
            case IRConst() | int():
                pass
            case BlockPlace(block=block, index=index, offset=_):
                self.update_uses(block, uses)
                self.update_uses(index, uses)
            case TempBlock() | SSAPlace():
                uses.add(stmt)
            case _:
                raise TypeError(f"Unexpected statement type: {type(stmt)}")
        return uses


class AdvancedDeadCodeElimination(CompilerPass):
    """Slower than regular DeadCodeElimination but can handle cases like definitions after the last use and so on."""

    def requires(self) -> set[CompilerPass]:
        return {LivenessAnalysis()}

    def run(self, entry: BasicBlock, config: OptimizerConfig) -> BasicBlock:
        for block in traverse_cfg_preorder(entry):
            live_stmts = []
            for statement in block.statements:
                live = get_live(statement)
                match statement:
                    case IRSet(place=place, value=value):
                        is_live = not (
                            (isinstance(place, SSAPlace) and place not in live)
                            or (
                                isinstance(place, BlockPlace)
                                and isinstance(place.block, TempBlock)
                                and place.block not in live
                            )
                            or (isinstance(value, IRGet) and place == value.place)
                        )
                        if is_live:
                            live_stmts.append(statement)
                        elif isinstance(value, IRInstr) and value.op.side_effects:
                            live_stmts.append(value)
                            value.live = live
                    case other:
                        live_stmts.append(other)
            block.statements = live_stmts
            block.phis = {place: phi for place, phi in block.phis.items() if place in get_live_phi_targets(block)}
        return entry



from collections.abc import Callable
from datetime import datetime
from pathlib import Path
from typing import cast

from sonolus.build.collection import Asset, Collection, Srl
from sonolus.build.compile import CompileCache
from sonolus.build.engine import package_engine, unpackage_data
from sonolus.build.level import package_level_data
from sonolus.script.engine import Engine
from sonolus.script.internal.context import ProjectContextState
from sonolus.script.level import ExternalLevelData, ExternalLevelDataDict, Level, LevelData, parse_external_level_data
from sonolus.script.project import BuildConfig, Project, ProjectSchema

BLANK_PNG = (
    b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x01\x00\x00\x00\x007n\xf9$"
    b"\x00\x00\x00\nIDATx\x01c`\x00\x00\x00\x02\x00\x01su\x01\x18\x00\x00\x00\x00IEND\xaeB`\x82"
)
BLANK_AUDIO = (
    b"RIFF$\x00\x00\x00WAVEfmt \x10\x00\x00\x00\x01\x00\x01\x00D\xac\x00\x00\x88X\x01\x00\x02"
    b"\x00\x10\x00data\x00\x00\x00\x00"
)


def build_project_to_collection(
    project: Project,
    config: BuildConfig | None,
    cache: CompileCache | None = None,
    project_state: ProjectContextState | None = None,
) -> Collection:
    collection = load_resources_files_to_collection(project.resources)
    build_project_to_existing_collection(project, collection, config, cache=cache, project_state=project_state)
    return collection


def build_project_to_existing_collection(
    project: Project,
    collection: Collection,
    config: BuildConfig | None,
    cache: CompileCache | None = None,
    project_state: ProjectContextState | None = None,
) -> None:
    for src_engine, converter in project.converters.items():
        if src_engine is None:
            continue
        apply_converter_to_collection(collection, src_engine, converter)
    fallback_converter = project.converters.get(None)
    if fallback_converter is not None:
        apply_converter_to_collection(collection, None, fallback_converter)
    if config.override_resource_level_engines:
        for level in collection.categories.get("levels", {}).values():
            level["item"]["engine"] = project.engine.name
    add_engine_to_collection(collection, project, project.engine, config, cache=cache, project_state=project_state)
    for level in project.levels:
        add_level_to_collection(collection, project, level)
    collection.name = f"{project.engine.name}"


def path_was_modified_after(path: Path, modified_time: datetime | int | float) -> bool:
    if isinstance(modified_time, datetime):
        target_timestamp = modified_time.timestamp()
    else:
        target_timestamp = float(modified_time)

    if path.stat().st_mtime > target_timestamp:
        return True

    if path.is_dir():
        for child in path.rglob("*"):
            if child.stat().st_mtime > target_timestamp:
                return True

    return False


def apply_converter_to_collection(
    self, src_engine: str | None, converter: Callable[[ExternalLevelData], LevelData | None]
) -> None:
    for level_details in self.categories.get("levels", {}).values():
        level = level_details["item"]
        if (
            src_engine is not None
            and level["engine"] != src_engine
            and not (isinstance(level["engine"], dict) and level["engine"].get("name") == src_engine)
        ):
            continue
        packaged_data_srl = level.get("data")
        packaged_data = self.repository.get(packaged_data_srl["hash"])
        data = unpackage_data(packaged_data)
        if not (isinstance(data, dict) and "bgmOffset" in data and "entities" in data):
            raise ValueError(f"Level data for level '{level['name']}' is not valid")
        parsed_data = parse_external_level_data(cast(ExternalLevelDataDict, data))
        new_data = converter(parsed_data)
        if new_data is None:
            continue
        packaged_new_data = package_level_data(new_data)
        new_data_srl = self.add_asset(packaged_new_data)
        level["data"] = new_data_srl


def add_engine_to_collection(
    collection: Collection,
    project: Project,
    engine: Engine,
    config: BuildConfig | None,
    cache: CompileCache | None = None,
    project_state: ProjectContextState | None = None,
):
    packaged_engine = package_engine(engine.data, config, cache=cache, project_state=project_state)
    item = {
        "name": engine.name,
        "version": engine.version,
        "title": engine.title,
        "subtitle": engine.subtitle,
        "author": engine.author,
        "tags": [tag.as_dict() for tag in engine.tags],
        "skin": collection.get_item("skins", engine.skin) if engine.skin else collection.get_default_item("skins"),
        "background": collection.get_item("backgrounds", engine.background)
        if engine.background
        else collection.get_default_item("backgrounds"),
        "effect": collection.get_item("effects", engine.effect)
        if engine.effect
        else collection.get_default_item("effects"),
        "particle": collection.get_item("particles", engine.particle)
        if engine.particle
        else collection.get_default_item("particles"),
        "thumbnail": load_resource(collection, engine.thumbnail, project.resources, BLANK_PNG),
        "playData": collection.add_asset(packaged_engine.play_data),
        "watchData": collection.add_asset(packaged_engine.watch_data),
        "previewData": collection.add_asset(packaged_engine.preview_data),
        "tutorialData": collection.add_asset(packaged_engine.tutorial_data),
        "rom": collection.add_asset(packaged_engine.rom),
        "configuration": collection.add_asset(packaged_engine.configuration),
    }
    if engine.description is not None:
        item["description"] = engine.description
    collection.add_item("engines", engine.name, item)


def add_level_to_collection(collection: Collection, project: Project, level: Level):
    packaged_level_data = package_level_data(level.data)
    item = {
        "name": level.name,
        "version": level.version,
        "rating": level.rating,
        "title": level.title,
        "artists": level.artists,
        "author": level.author,
        "tags": [tag.as_dict() for tag in level.tags],
        "engine": collection.get_item("engines", project.engine.name),
        "useSkin": {"useDefault": True}
        if level.use_skin is None
        else {"useDefault": False, "item": collection.get_item("skins", level.use_skin)},
        "useBackground": {"useDefault": True}
        if level.use_background is None
        else {"useDefault": False, "item": collection.get_item("backgrounds", level.use_background)},
        "useEffect": {"useDefault": True}
        if level.use_effect is None
        else {"useDefault": False, "item": collection.get_item("effects", level.use_effect)},
        "useParticle": {"useDefault": True}
        if level.use_particle is None
        else {"useDefault": False, "item": collection.get_item("particles", level.use_particle)},
        "cover": load_resource(collection, level.cover, project.resources, BLANK_PNG),
        "bgm": load_resource(collection, level.bgm, project.resources, BLANK_AUDIO),
        "data": collection.add_asset(packaged_level_data),
    }
    if level.description is not None:
        item["description"] = level.description
    if level.preview is not None:
        item["preview"] = load_resource(collection, level.preview, project.resources, BLANK_AUDIO)
    collection.add_item("levels", level.name, item)


def load_resource(collection: Collection, asset: Asset | None, base_path: Path, default: bytes) -> Srl:
    if asset is None:
        return collection.add_asset(default)
    if isinstance(asset, str) and not asset.startswith(("http://", "https://")):
        return collection.add_asset(base_path / asset)
    return collection.add_asset(asset)


def load_resources_files_to_collection(base_path: Path) -> Collection:
    collection = Collection()
    for path in base_path.rglob("*.scp"):
        collection.load_from_scp(path)
    collection.load_from_source(base_path)
    return collection


def get_project_schema(project: Project) -> ProjectSchema:
    by_archetype: dict[str, dict[str, bool]] = {}
    for archetype in project.engine.data.play.archetypes:
        archetype._init_fields()
        fields = by_archetype.setdefault(archetype.name, {})
        # If a field is exported, we should exclude it if it's imported in watch mode
        for field in archetype._exported_keys_:
            fields[field] = False
        for field in archetype._imported_keys_:
            fields[field] = True
    for archetype in project.engine.data.watch.archetypes:
        archetype._init_fields()
        fields = by_archetype.setdefault(archetype.name, {})
        for field in archetype._imported_keys_:
            if field in {"#ACCURACY", "#JUDGMENT"}:
                continue
            if field not in fields:
                fields[field] = True
    for archetype in project.engine.data.preview.archetypes:
        archetype._init_fields()
        fields = by_archetype.setdefault(archetype.name, {})
        for field in archetype._imported_keys_:
            fields[field] = True
    return {
        "archetypes": [
            {
                "name": name,
                "fields": [*fields],
            }
            for name, fields in by_archetype.items()
        ]
    }



from __future__ import annotations

import ast
import builtins
import functools
import inspect
from collections.abc import Callable, Iterable, Sequence
from inspect import ismethod
from time import perf_counter_ns
from types import FunctionType, MethodType, MethodWrapperType
from typing import Any, Never

from sonolus.backend.excepthook import install_excepthook
from sonolus.backend.utils import get_function, scan_writes
from sonolus.script.debug import assert_true
from sonolus.script.internal.builtin_impls import BUILTIN_IMPLS, _bool, _float, _int, _len, _super
from sonolus.script.internal.constant import ConstantValue
from sonolus.script.internal.context import (
    ConflictBinding,
    Context,
    EmptyBinding,
    FunctionVisitStatistics,
    Scope,
    ValueBinding,
    ctx,
    set_ctx,
    using_ctx,
)
from sonolus.script.internal.descriptor import SonolusDescriptor
from sonolus.script.internal.error import CompilationError
from sonolus.script.internal.impl import validate_value
from sonolus.script.internal.meta_fn import meta_fn
from sonolus.script.internal.transient import TransientValue
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.internal.value import Value
from sonolus.script.iterator import SonolusIterator
from sonolus.script.maybe import Maybe, Nothing
from sonolus.script.num import Num, _is_num
from sonolus.script.record import Record

_compiler_internal_ = True


def compile_and_call[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> R:
    if not ctx():
        return fn(*args, **kwargs)
    if getattr(fn, "_meta_fn_", False):
        return validate_value(fn(*args, **kwargs))
    return validate_value(generate_fn_impl(fn)(*args, **kwargs))


def compile_and_call_at_definition[**P, R](fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs) -> R:
    if not ctx():
        return fn(*args, **kwargs)
    if getattr(fn, "_meta_fn_", False):
        return validate_value(fn(*args, **kwargs))
    source_file, node = get_function(fn)
    if ctx().no_eval:
        debug_stack = ctx().callback_state.debug_stack
        try:
            debug_stack.append(f'File "{source_file}", line {node.lineno}, in <callback>')
            return compile_and_call(fn, *args, **kwargs)
        finally:
            debug_stack.pop()
    location_args = {
        "lineno": node.lineno,
        "col_offset": node.col_offset,
        "end_lineno": node.lineno,
        "end_col_offset": node.col_offset + 1,
    }
    expr = ast.Expression(
        body=ast.Call(
            func=ast.Name(id="fn", ctx=ast.Load(), **location_args),
            args=[],
            keywords=[],
            **location_args,
        )
    )

    debug_stack = ctx().callback_state.debug_stack
    try:
        debug_stack.append(f'File "{source_file}", line {node.lineno}, in <callback>')
        return eval(
            compile(expr, filename=source_file, mode="eval").replace(co_name="<callback>"),
            {"fn": lambda: compile_and_call(fn, *args, **kwargs), "_filter_traceback_": True, "_traceback_root_": True},
        )
    finally:
        debug_stack.pop()


def generate_fn_impl(fn: Callable):
    install_excepthook()
    match fn:
        case ConstantValue() as value if value._is_py_():
            return generate_fn_impl(value._as_py_())
        case MethodType() as method:
            return functools.partial(generate_fn_impl(method.__func__), method.__self__)
        case FunctionType() as function:
            if getattr(function, "_meta_fn_", False):
                return function
            return functools.partial(eval_fn, function)
        case _:
            if callable(fn) and (isinstance(fn, Value) or getattr(fn, "_is_comptime_value_", False)):
                return generate_fn_impl(fn.__call__)
            elif callable(fn):
                raise TypeError(f"Unsupported callable {fn!r}")
            else:
                raise TypeError(f"'{type(fn).__name__}' object is not callable")


def eval_fn(fn: Callable, /, *args, **kwargs):
    source_file, node = get_function(fn)
    bound_args = inspect.signature(fn).bind(*args, **kwargs)
    bound_args.apply_defaults()
    if ismethod(fn):
        code = fn.__func__.__code__
        closure = fn.__func__.__closure__
    else:
        code = fn.__code__
        closure = fn.__closure__
    if closure is None:
        nonlocal_vars = {}
    else:
        nonlocal_vars = {
            var: cell.cell_contents for var, cell in zip(code.co_freevars, closure, strict=True) if cell is not None
        }
    global_vars = {
        **builtins.__dict__,
        **fn.__globals__,
        **nonlocal_vars,
    }
    function_name = getattr(fn, "__name__", "<unnamed>")
    module = getattr(fn, "__module__", None)
    if module is not None:
        qualified_name = f"{module}.{getattr(fn, '__qualname__', function_name)}"
    else:
        qualified_name = f"<unknown>.{getattr(fn, '__qualname__', function_name)}"
    return Visitor(
        source_file, bound_args, global_vars, parent=None, function_name=function_name, qualified_name=qualified_name
    ).run(node)


unary_ops = {
    ast.Invert: "__invert__",
    ast.UAdd: "__pos__",
    ast.USub: "__neg__",
}

bin_ops = {
    ast.Add: "__add__",
    ast.Sub: "__sub__",
    ast.Mult: "__mul__",
    ast.Div: "__truediv__",
    ast.FloorDiv: "__floordiv__",
    ast.Mod: "__mod__",
    ast.Pow: "__pow__",
    ast.LShift: "__lshift__",
    ast.RShift: "__rshift__",
    ast.BitOr: "__or__",
    ast.BitAnd: "__and__",
    ast.BitXor: "__xor__",
    ast.MatMult: "__matmul__",
}

rbin_ops = {
    ast.Add: "__radd__",
    ast.Sub: "__rsub__",
    ast.Mult: "__rmul__",
    ast.Div: "__rtruediv__",
    ast.FloorDiv: "__rfloordiv__",
    ast.Mod: "__rmod__",
    ast.Pow: "__rpow__",
    ast.LShift: "__rlshift__",
    ast.RShift: "__rrshift__",
    ast.BitOr: "__ror__",
    ast.BitAnd: "__rand__",
    ast.BitXor: "__rxor__",
    ast.MatMult: "__rmatmul__",
}

inplace_ops = {
    ast.Add: "__iadd__",
    ast.Sub: "__isub__",
    ast.Mult: "__imul__",
    ast.Div: "__itruediv__",
    ast.FloorDiv: "__ifloordiv__",
    ast.Mod: "__imod__",
    ast.Pow: "__ipow__",
    ast.LShift: "__ilshift__",
    ast.RShift: "__irshift__",
    ast.BitOr: "__ior__",
    ast.BitXor: "__ixor__",
    ast.BitAnd: "__iand__",
    ast.MatMult: "__imatmul__",
}

comp_ops = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__lt__",
    ast.LtE: "__le__",
    ast.Gt: "__gt__",
    ast.GtE: "__ge__",
}

rcomp_ops = {
    ast.Eq: "__eq__",
    ast.NotEq: "__ne__",
    ast.Lt: "__gt__",
    ast.LtE: "__ge__",
    ast.Gt: "__lt__",
    ast.GtE: "__le__",
    ast.In: "__contains__",  # Only supported on the right side
    ast.NotIn: "__contains__",
}

op_to_symbol = {
    ast.Add: "+",
    ast.Sub: "-",
    ast.Mult: "*",
    ast.Div: "/",
    ast.FloorDiv: "//",
    ast.Mod: "%",
    ast.Pow: "**",
    ast.Eq: "==",
    ast.NotEq: "!=",
    ast.Lt: "<",
    ast.LtE: "<=",
    ast.Gt: ">",
    ast.GtE: ">=",
    ast.And: "and",
    ast.Or: "or",
    ast.BitAnd: "&",
    ast.BitOr: "|",
    ast.BitXor: "^",
    ast.LShift: "<<",
    ast.RShift: ">>",
    ast.USub: "-",
    ast.UAdd: "+",
    ast.Invert: "~",
    ast.Not: "not",
    ast.In: "in",
    ast.NotIn: "not in",
}


def record_visit_time(function_name: str, total_time: int, own_time: int) -> None:
    visit_stats = ctx().project_state.visit_stats
    if function_name not in visit_stats:
        visit_stats[function_name] = stats = FunctionVisitStatistics()
    else:
        stats = visit_stats[function_name]
    # Not accurate if multithreaded, but it's preferable to collect performance statistics single-threaded anyway
    stats.total_time += total_time
    stats.own_time += own_time
    stats.call_count += 1
    ctx().callback_state.visitor_own_time += own_time


def mark_start(function_name: str) -> Callable[[], None]:
    start_time = perf_counter_ns()
    start_own_time = ctx().callback_state.visitor_own_time

    def mark_end():
        total_time = perf_counter_ns() - start_time
        record_visit_time(
            function_name, total_time, total_time - (ctx().callback_state.visitor_own_time - start_own_time)
        )

    return mark_end


class Visitor(ast.NodeVisitor):
    source_file: str
    globals: dict[str, Any]
    bound_args: inspect.BoundArguments
    used_names: dict[str, int]
    return_ctxs: list[Context]  # Contexts at return statements, which will branch to the exit
    loop_head_ctxs: list[
        Context | list[Context]
    ]  # Contexts at loop heads, from outer to inner. Contains a list for unrolled (tuple) loops
    break_ctxs: list[list[Context]]  # Contexts at break statements, from outer to inner
    yield_ctxs: list[Context]  # Contexts at yield statements, which will branch to the exit
    resume_ctxs: list[Context]  # Contexts after yield statements
    active_ctx: Context | None  # The active context for use in nested functions
    parent: Visitor | None  # The parent visitor for use in nested functions
    used_parent_binding_values: dict[str, Value]  # Values of parent bindings used in this
    function_name: str
    qualified_name: str

    def __init__(
        self,
        source_file: str,
        bound_args: inspect.BoundArguments,
        global_vars: dict[str, Any],
        parent: Visitor | None,
        function_name: str,
        qualified_name: str | None = None,
    ):
        self.source_file = source_file
        self.globals = global_vars
        self.bound_args = bound_args
        self.used_names = {}
        self.return_ctxs = []
        self.loop_head_ctxs = []
        self.break_ctxs = []
        self.yield_ctxs = []
        self.resume_ctxs = []
        self.active_ctx = None
        self.parent = parent
        self.used_parent_binding_values = {}
        self.function_name = function_name
        if qualified_name is None:
            if parent is None:
                self.qualified_name = function_name
            else:
                self.qualified_name = f"{parent.qualified_name}.{function_name}"
        else:
            self.qualified_name = qualified_name

    def run(self, node):
        completion_timer = mark_start(self.qualified_name)
        before_ctx = ctx()
        before_alloc_state = ctx().save_alloc_state()
        start_ctx = before_ctx.branch_with_scope(None, Scope())
        set_ctx(start_ctx)
        for name, value in self.bound_args.arguments.items():
            ctx().scope.set_value(name, validate_value(value))
        match node:
            case ast.FunctionDef(body=body):
                ctx().scope.set_value("$return", validate_value(None))
                for stmt in body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            case ast.Lambda(body=body):
                result = self.visit(body)
                ctx().scope.set_value("$return", result)
            case ast.GeneratorExp(elt=elt, generators=generators):
                first_generator = generators[0]
                iterable = self.visit(first_generator.iter)
                if isinstance(iterable, TupleImpl):
                    initial_iterator = iterable
                else:
                    if not hasattr(iterable, "__iter__"):
                        raise TypeError(f"Object of type '{type(iterable).__name__}' is not iterable")
                    initial_iterator = self.handle_call(first_generator.iter, iterable.__iter__)
                    if not isinstance(initial_iterator, SonolusIterator):
                        raise ValueError("Unsupported iterator")
                # The initial iterator is evaluated eagerly in Python
                before_ctx = ctx().branch_with_scope(None, before_ctx.scope.copy())
                start_ctx = before_ctx.branch_with_scope(None, Scope())
                set_ctx(start_ctx)
                self.construct_genexpr(generators, elt, initial_iterator)
                ctx().scope.set_value("$return", validate_value(None))
            case _:
                raise NotImplementedError("Unsupported syntax")
        # has_yield is set by the find_function visitor
        if getattr(node, "has_yield", False) or isinstance(node, ast.GeneratorExp):
            return_ctx = Context.meet([*self.return_ctxs, ctx()])
            result_binding = return_ctx.scope.get_binding("$return")
            if not isinstance(result_binding, ValueBinding):
                raise ValueError("Function has conflicting return values")
            if not result_binding.value._is_py_() and result_binding.value._as_py_() is not None:
                raise ValueError("Generator function return statements must return None")
            with using_ctx(start_ctx):
                state_var = Num._alloc_()
                is_present_var = Num._alloc_()
            with using_ctx(before_ctx):
                state_var._set_(0)
            with using_ctx(return_ctx):
                is_present_var._set_(0)
            del before_ctx.outgoing[None]  # Unlink the state machine body from the call site
            entry = before_ctx.new_empty_disconnected()
            entry.test = state_var.ir()
            for i, tgt in enumerate([start_ctx, *self.resume_ctxs]):
                entry.outgoing[i] = tgt
            entry.outgoing[None] = return_ctx
            yield_between_ctxs = []
            for i, out in enumerate(self.yield_ctxs, start=1):
                between = out.branch(None)
                with using_ctx(between):
                    state_var._set_(i)
                yield_between_ctxs.append(between)
            if yield_between_ctxs:
                yield_merge_ctx = Context.meet(yield_between_ctxs)
            else:
                yield_merge_ctx = before_ctx.new_empty_disconnected()
            yield_binding = yield_merge_ctx.scope.get_binding("$yield")
            match yield_binding:
                case ValueBinding():
                    with using_ctx(yield_merge_ctx):
                        is_present_var._set_(1)
                    yield_value = Maybe(present=is_present_var, value=yield_binding.value)
                case EmptyBinding():
                    yield_value = Nothing
                case ConflictBinding():
                    raise ValueError("Function has conflicting yield values")
            next_result_ctx = Context.meet([yield_merge_ctx, return_ctx])
            set_ctx(before_ctx)
            return_test = Num._alloc_()
            next_result_ctx.test = return_test.ir()
            completion_timer()
            return Generator(
                return_test,
                entry,
                next_result_ctx,
                yield_value,
                self.used_parent_binding_values,
                self,
            )
        after_ctx = Context.meet([*self.return_ctxs, ctx()])
        self.active_ctx = after_ctx
        result_binding = after_ctx.scope.get_binding("$return")
        if not isinstance(result_binding, ValueBinding):
            raise ValueError("Function has conflicting return values")
        set_ctx(after_ctx.branch_with_scope(None, before_ctx.scope.copy()))
        # Noting could have escaped, so allow reuse, which can allow naive allocation to succeed in the optimizer for
        # better compile times.
        if result_binding.value is validate_value(None):
            ctx().restore_alloc_state(before_alloc_state)
        completion_timer()
        return result_binding.value

    def construct_genexpr(
        self, generators: Iterable[ast.comprehension], elt: ast.expr, initial_iterator: Value | None = None
    ):
        if not generators:
            # Note that there may effectively be multiple yields in an expression since
            # tuples are unrolled.
            value = self.visit(elt)
            ctx().scope.set_value("$yield", validate_value(value))
            self.yield_ctxs.append(ctx())
            resume_ctx = ctx().new_disconnected()
            self.resume_ctxs.append(resume_ctx)
            set_ctx(resume_ctx)
            return
        generator, *others = generators
        if initial_iterator is not None:
            iterable = initial_iterator
        else:
            iterable = self.visit(generator.iter)
        if isinstance(iterable, TupleImpl):
            for value in iterable.value:
                set_ctx(ctx().branch(None))
                self.handle_assign(generator.target, validate_value(value))
                self.construct_genexpr(others, elt)
        else:
            if initial_iterator is not None:
                iterator = initial_iterator
            else:
                if not hasattr(iterable, "__iter__"):
                    raise TypeError(f"Object of type '{type(iterable).__name__}' is not iterable")
                iterator = self.handle_call(generator.iter, iterable.__iter__)
            if not isinstance(iterator, SonolusIterator):
                raise ValueError("Unsupported iterator")
            header_ctx = ctx().branch(None)
            set_ctx(header_ctx)
            next_value = self.handle_call(generator.iter, iterator.next)
            if not isinstance(next_value, Maybe):
                raise ValueError("Iterator next must return a Maybe")
            if next_value._present._is_py_() and not next_value._present._as_py_():
                # This will never run
                return
            ctx().test = next_value._present.ir()
            body_ctx = ctx().branch(None)
            else_ctx = ctx().branch(0)
            set_ctx(body_ctx)
            self.handle_assign(generator.target, next_value._value)
            for if_expr in generator.ifs:
                test = self.convert_to_boolean_num(if_expr, self.visit(if_expr))
                if test._is_py_():
                    if test._as_py_():
                        continue
                    else:
                        ctx().outgoing[None] = header_ctx
                        set_ctx(ctx().into_dead())
                else:
                    if_then_ctx = ctx().branch(None)
                    if_else_ctx = ctx().branch(0)
                    ctx().test = test.ir()
                    if_else_ctx.outgoing[None] = header_ctx
                    set_ctx(if_then_ctx)
            self.construct_genexpr(others, elt)
            ctx().outgoing[None] = header_ctx
            set_ctx(else_ctx)

    def visit(self, node):
        """Visit a node."""
        # We want this here so this is filtered out of tracebacks
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        # with self.reporting_errors_at_node(node):
        #     return visitor(node)
        try:
            return visitor(node)
        except Exception as e:
            if isinstance(e, CompilationError):
                raise
            self.raise_exception_at_node(node, e)

    def visit_FunctionDef(self, node):
        name = node.name
        signature = self.arguments_to_signature(node.args)

        def fn(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            return Visitor(
                self.source_file,
                bound,
                self.globals,
                parent=self,
                function_name=name,
            ).run(node)

        fn._meta_fn_ = True
        fn.__name__ = name
        fn.__qualname__ = name

        for decorator in reversed(node.decorator_list):
            fn = self.handle_call(decorator, self.visit(decorator), fn)

        ctx().scope.set_value(name, validate_value(fn))

    def visit_AsyncFunctionDef(self, node):
        raise NotImplementedError("Async functions are not supported")

    def visit_ClassDef(self, node):
        raise NotImplementedError("Classes within functions are not supported")

    def visit_Return(self, node):
        value = self.visit(node.value) if node.value else validate_value(None)
        ctx().scope.set_value("$return", value)
        self.return_ctxs.append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Delete(self, node):
        for target in node.targets:
            match target:
                case ast.Name():
                    raise NotImplementedError("Deleting variables is not supported")
                case ast.Subscript(value=value, slice=slice):
                    self.handle_delitem(target, self.visit(value), self.visit(slice))
                case ast.Attribute():
                    raise NotImplementedError("Deleting attributes is not supported")
                case _:
                    raise NotImplementedError("Unsupported delete target")

    def visit_Assign(self, node):
        value = self.visit(node.value)
        for target in node.targets:
            self.handle_assign(target, value)

    def visit_TypeAlias(self, node):
        raise NotImplementedError("Type aliases are not supported")

    def visit_AugAssign(self, node):
        lhs_value = self.visit(node.target)
        rhs_value = self.visit(node.value)
        inplace_fn_name = inplace_ops[type(node.op)]
        regular_fn_name = bin_ops[type(node.op)]
        right_fn_name = rbin_ops[type(node.op)]
        if hasattr(lhs_value, inplace_fn_name):
            result = self.handle_call(node, getattr(lhs_value, inplace_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                if result is not lhs_value:
                    raise ValueError("Inplace operation must return the same object")
                self.handle_assign(node.target, result)
                return
        if hasattr(lhs_value, regular_fn_name):
            result = self.handle_call(node, getattr(lhs_value, regular_fn_name), rhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        if hasattr(rhs_value, right_fn_name) and type(lhs_value) is not type(rhs_value):
            result = self.handle_call(node, getattr(rhs_value, right_fn_name), lhs_value)
            if not self.is_not_implemented(result):
                self.handle_assign(node.target, result)
                return
        raise TypeError(
            f"unsupported operand type(s) for {op_to_symbol[type(node.op)]}=: "
            f"'{type(lhs_value).__name__}' and '{type(rhs_value).__name__}'"
        )

    def visit_AnnAssign(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)

    def visit_For(self, node):
        from sonolus.script.internal.tuple_impl import TupleImpl

        iterable = self.visit(node.iter)
        if isinstance(iterable, TupleImpl):
            # Unroll the loop
            break_ctxs = []
            for value in iterable.value:
                set_ctx(ctx().branch(None))
                self.loop_head_ctxs.append([])
                self.break_ctxs.append([])
                self.handle_assign(node.target, validate_value(value))
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                continue_ctxs = [*self.loop_head_ctxs.pop(), ctx()]
                break_ctxs.extend(self.break_ctxs.pop())
                set_ctx(Context.meet(continue_ctxs))
            if break_ctxs:
                set_ctx(Context.meet([*break_ctxs, ctx()]))
            return
        if not hasattr(iterable, "__iter__"):
            raise TypeError(f"Object of type '{type(iterable).__name__}' is not iterable")
        iterator = self.handle_call(node, iterable.__iter__)
        if not isinstance(iterator, SonolusIterator):
            raise ValueError("Unsupported iterator")
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        next_value = self.handle_call(node, iterator.next)
        if not isinstance(next_value, Maybe):
            raise ValueError("Iterator next must return a Maybe")
        if next_value._present._is_py_() and not next_value._present._as_py_():
            # The loop will never run, continue after evaluating the condition
            self.loop_head_ctxs.pop()
            self.break_ctxs.pop()
            for stmt in node.orelse:
                if not ctx().live:
                    break
                self.visit(stmt)
            return
        ctx().test = next_value._present.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        self.handle_assign(node.target, next_value._value)
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        self.loop_head_ctxs.pop()
        break_ctxs = self.break_ctxs.pop()

        set_ctx(else_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        else_end_ctx = ctx()

        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_While(self, node):
        writes = scan_writes(node)
        header_ctx = ctx().prepare_loop_header(writes)
        self.loop_head_ctxs.append(header_ctx)
        self.break_ctxs.append([])
        set_ctx(header_ctx)
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))
        if test._is_py_():
            if test._as_py_():
                # The loop will run until a break / return
                body_ctx = ctx().branch(None)
                set_ctx(body_ctx)
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                ctx().branch_to_loop_header(header_ctx)

                self.loop_head_ctxs.pop()
                break_ctxs = self.break_ctxs.pop()

                # Skip the else block

                after_ctx = Context.meet([ctx().into_dead(), *break_ctxs])
                set_ctx(after_ctx)
                return
            else:
                # The loop will never run, continue after evaluating the condition
                self.loop_head_ctxs.pop()
                self.break_ctxs.pop()
                for stmt in node.orelse:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                return
        ctx().test = test.ir()
        body_ctx = ctx().branch(None)
        else_ctx = ctx().branch(0)

        set_ctx(body_ctx)
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        ctx().branch_to_loop_header(header_ctx)

        self.loop_head_ctxs.pop()
        break_ctxs = self.break_ctxs.pop()

        set_ctx(else_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        else_end_ctx = ctx()

        after_ctx = Context.meet([else_end_ctx, *break_ctxs])
        set_ctx(after_ctx)

    def visit_If(self, node):
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                for stmt in node.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            else:
                for stmt in node.orelse:
                    if not ctx().live:
                        break
                    self.visit(stmt)
            return

        ctx_init = ctx()
        ctx_init.test = test.ir()
        true_ctx = ctx_init.branch(None)
        false_ctx = ctx_init.branch(0)

        set_ctx(true_ctx)
        for stmt in node.body:
            if not ctx().live:
                break
            self.visit(stmt)
        true_end_ctx = ctx()

        set_ctx(false_ctx)
        for stmt in node.orelse:
            if not ctx().live:
                break
            self.visit(stmt)
        false_end_ctx = ctx()

        set_ctx(Context.meet([true_end_ctx, false_end_ctx]))

    def visit_With(self, node):
        raise NotImplementedError("With statements are not supported")

    def visit_AsyncWith(self, node):
        raise NotImplementedError("Async with statements are not supported")

    def visit_Match(self, node):
        subject = self.visit(node.subject)
        end_ctxs = []
        for case in node.cases:
            if not ctx().live:
                break
            true_ctx, false_ctx = self.handle_match_pattern(subject, case.pattern)
            if not true_ctx.live:
                set_ctx(false_ctx)
                continue
            set_ctx(true_ctx)
            guard = (
                self.convert_to_boolean_num(case.guard, self.visit(case.guard)) if case.guard else validate_value(True)
            )
            if guard._is_py_():
                if guard._as_py_():
                    for stmt in case.body:
                        if not ctx().live:
                            break
                        self.visit(stmt)
                    end_ctxs.append(ctx())
                else:
                    # Merge failing before the guard and failing now at the guard (which we know is guaranteed to fail)
                    false_ctx = Context.meet([ctx(), false_ctx])
            else:
                ctx().test = guard.ir()
                guard_true_ctx = ctx().branch(None)
                guard_false_ctx = ctx().branch(0)
                set_ctx(guard_true_ctx)
                for stmt in case.body:
                    if not ctx().live:
                        break
                    self.visit(stmt)
                end_ctxs.append(ctx())
                false_ctx = Context.meet([false_ctx, guard_false_ctx])
            set_ctx(false_ctx)
        end_ctxs.append(ctx())
        if end_ctxs:
            set_ctx(Context.meet(end_ctxs))

    def handle_match_pattern(self, subject: Value, pattern: ast.pattern) -> tuple[Context, Context]:
        from sonolus.script.internal.generic import validate_type_spec
        from sonolus.script.internal.tuple_impl import TupleImpl

        if not ctx().live:
            return ctx().into_dead(), ctx()

        match pattern:
            case ast.MatchValue(value=value):
                value = self.visit(value)
                test = self.convert_to_boolean_num(pattern, validate_value(subject == value))
                if test._is_py_():
                    if test._as_py_():
                        return ctx(), ctx().into_dead()
                    else:
                        return ctx().into_dead(), ctx()
                ctx_init = ctx()
                ctx_init.test = test.ir()
                true_ctx = ctx_init.branch(None)
                false_ctx = ctx_init.branch(0)
                return true_ctx, false_ctx
            case ast.MatchSingleton(value=value):
                match value:
                    case True:
                        raise NotImplementedError("Matching against True is not supported, use 1 instead")
                    case False:
                        raise NotImplementedError("Matching against False is not supported, use 0 instead")
                    case None:
                        test = validate_value(subject._is_py_() and subject._as_py_() is None)
                    case _:
                        raise NotImplementedError("Unsupported match singleton")
                if test._is_py_():
                    if test._as_py_():
                        return ctx(), ctx().into_dead()
                    else:
                        return ctx().into_dead(), ctx()
                else:
                    ctx_init = ctx()
                    ctx_init.test = test.ir()
                    true_ctx = ctx_init.branch(None)
                    false_ctx = ctx_init.branch(0)
                    return true_ctx, false_ctx
            case ast.MatchSequence(patterns=patterns):
                target_len = len(patterns)
                if not (isinstance(subject, Sequence | TupleImpl)):
                    return ctx().into_dead(), ctx()
                length_test = self.convert_to_boolean_num(pattern, validate_value(_len(subject) == target_len))
                ctx_init = ctx()
                if not length_test._is_py_():
                    ctx_init.test = length_test.ir()
                    true_ctx = ctx_init.branch(None)
                    false_ctxs = [ctx_init.branch(0)]
                elif length_test._as_py_():
                    true_ctx = ctx_init
                    false_ctxs = []
                else:
                    return ctx().into_dead(), ctx()
                set_ctx(true_ctx)
                for i, subpattern in enumerate(patterns):
                    if not ctx().live:
                        break
                    value = self.handle_getitem(subpattern, subject, validate_value(i))
                    true_ctx, false_ctx = self.handle_match_pattern(value, subpattern)
                    false_ctxs.append(false_ctx)
                    set_ctx(true_ctx)
                return true_ctx, Context.meet(false_ctxs)
            case ast.MatchMapping():
                raise NotImplementedError("Match mappings are not supported")
            case ast.MatchClass(cls=cls, patterns=patterns, kwd_attrs=kwd_attrs, kwd_patterns=kwd_patterns):
                cls = self.visit(cls)
                if cls._is_py_() and cls._as_py_() in {_int, _float, _bool}:
                    raise TypeError("Instance check against int, float, or bool is not supported, use Num instead")
                cls = validate_type_spec(cls)
                if not isinstance(cls, type):
                    raise TypeError("Class is not a type")
                if not isinstance(subject, cls):
                    return ctx().into_dead(), ctx()
                if patterns:
                    if not hasattr(cls, "__match_args__"):
                        raise TypeError("Class does not support match patterns")
                    if len(cls.__match_args__) < len(patterns):
                        raise ValueError("Too many match patterns")
                    # kwd_attrs can't be mixed with patterns on the syntax level,
                    # so we can just set it like this since it's empty
                    kwd_attrs = cls.__match_args__[: len(patterns)]
                    kwd_patterns = patterns
                if kwd_attrs:
                    true_ctx = ctx()
                    false_ctxs = []
                    for attr, subpattern in zip(kwd_attrs, kwd_patterns, strict=False):
                        if not hasattr(subject, attr):
                            raise AttributeError(f"Object has no attribute {attr}")
                        value = self.handle_getattr(subpattern, subject, attr)
                        true_ctx, false_ctx = self.handle_match_pattern(value, subpattern)
                        false_ctxs.append(false_ctx)
                        set_ctx(true_ctx)
                    return true_ctx, Context.meet(false_ctxs)
                return ctx(), ctx().into_dead()
            case ast.MatchStar():
                raise NotImplementedError("Match stars are not supported")
            case ast.MatchAs(pattern=pattern, name=name):
                if pattern:
                    true_ctx, false_ctx = self.handle_match_pattern(subject, pattern)
                    if name:
                        true_ctx.scope.set_value(name, subject)
                    return true_ctx, false_ctx
                else:
                    if name:
                        ctx().scope.set_value(name, subject)
                    return ctx(), ctx().into_dead()
            case ast.MatchOr():
                true_ctxs = []
                assert pattern.patterns
                for subpattern in pattern.patterns:
                    if not ctx().live:
                        break
                    true_ctx, false_ctx = self.handle_match_pattern(subject, subpattern)
                    true_ctxs.append(true_ctx)
                    set_ctx(false_ctx)
                return Context.meet(true_ctxs), ctx()

    def visit_Raise(self, node):
        raise NotImplementedError("Raise statements are not supported")

    def visit_Try(self, node):
        raise NotImplementedError("Try statements are not supported")

    def visit_TryStar(self, node):
        raise NotImplementedError("Try* statements are not supported")

    def visit_Assert(self, node):
        self.handle_call(
            node, assert_true, self.visit(node.test), self.visit(node.msg) if node.msg else validate_value(None)
        )

    def visit_Import(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_ImportFrom(self, node):
        raise NotImplementedError("Import statements are not supported")

    def visit_Global(self, node):
        raise NotImplementedError("Global statements are not supported")

    def visit_Nonlocal(self, node):
        raise NotImplementedError("Nonlocal statements are not supported")

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Pass(self, node):
        pass

    def visit_Break(self, node):
        self.break_ctxs[-1].append(ctx())
        set_ctx(ctx().into_dead())

    def visit_Continue(self, node):
        loop_head = self.loop_head_ctxs[-1]
        if isinstance(loop_head, list):
            loop_head.append(ctx())
        else:
            ctx().branch_to_loop_header(loop_head)
        set_ctx(ctx().into_dead())

    def visit_BoolOp(self, node) -> Value:
        match node.op:
            case ast.And():
                handler = self.handle_and
            case ast.Or():
                handler = self.handle_or
            case _:
                raise NotImplementedError(f"Unsupported bool operator {op_to_symbol[type(node.op)]}")

        if not node.values:
            raise ValueError("Bool operator requires at least one operand")
        if len(node.values) == 1:
            return self.visit(node.values[0])
        initial, *rest = node.values
        return handler(self.visit(initial), ast.copy_location(ast.BoolOp(op=node.op, values=rest), node))

    def visit_NamedExpr(self, node):
        value = self.visit(node.value)
        self.handle_assign(node.target, value)
        return value

    def visit_BinOp(self, node):
        lhs = self.visit(node.left)
        rhs = self.visit(node.right)
        op = bin_ops[type(node.op)]
        if lhs._is_py_() and rhs._is_py_():
            lhs_py = lhs._as_py_()
            rhs_py = rhs._as_py_()
            if (
                (isinstance(lhs_py, type) or getattr(lhs_py, "_is_comptime_value_", False))
                and (isinstance(rhs_py, type) or getattr(rhs_py, "_is_comptime_value_", False))
                and hasattr(type(lhs_py), op)
            ):
                return validate_value(getattr(lhs_py, op)(rhs_py))
        if hasattr(lhs, op):
            result = self.handle_call(node, getattr(lhs, op), rhs)
            if not self.is_not_implemented(result):
                return result
        if hasattr(rhs, rbin_ops[type(node.op)]) and type(lhs) is not type(rhs):
            result = self.handle_call(node, getattr(rhs, rbin_ops[type(node.op)]), lhs)
            if not self.is_not_implemented(result):
                return result
        raise TypeError(
            f"unsupported operand type(s) for {op_to_symbol[type(node.op)]}: "
            f"'{type(lhs).__name__}' and '{type(rhs).__name__}'"
        )

    def visit_UnaryOp(self, node):
        operand = self.visit(node.operand)
        if isinstance(node.op, ast.Not):
            return self.convert_to_boolean_num(node, operand).not_()
        op = unary_ops[type(node.op)]
        if operand._is_py_():
            operand_py = operand._as_py_()
            if isinstance(operand_py, type) and hasattr(type(operand_py), op):
                return self.handle_call(node, getattr(type(operand_py), op), operand_py)
        if hasattr(operand, op):
            return self.handle_call(node, getattr(operand, op))
        raise TypeError(f"bad operand type for unary {op_to_symbol[type(node.op)]}: '{type(operand).__name__}'")

    def visit_Lambda(self, node):
        signature = self.arguments_to_signature(node.args)

        def fn(*args, **kwargs):
            bound = signature.bind(*args, **kwargs)
            bound.apply_defaults()
            return Visitor(
                self.source_file,
                bound,
                self.globals,
                parent=self,
                function_name="<lambda>",
            ).run(node)

        fn._meta_fn_ = True
        fn.__name__ = "<lambda>"

        return validate_value(fn)

    def visit_IfExp(self, node):
        test = self.convert_to_boolean_num(node.test, self.visit(node.test))

        if test._is_py_():
            if test._as_py_():
                return self.visit(node.body)
            else:
                return self.visit(node.orelse)

        res_name = self.new_name("ifexp")
        ctx_init = ctx()
        ctx_init.test = test.ir()

        set_ctx(ctx_init.branch(None))
        true_value = self.visit(node.body)
        ctx().scope.set_value(res_name, true_value)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        false_value = self.visit(node.orelse)
        ctx().scope.set_value(res_name, false_value)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        return ctx().scope.get_value(res_name)

    def visit_Dict(self, node):
        return validate_value({self.visit(k): self.visit(v) for k, v in zip(node.keys, node.values, strict=True)})

    def visit_Set(self, node):
        from sonolus.script.containers import FrozenNumSet

        return self.handle_call(node, FrozenNumSet.of, *(self.visit(elt) for elt in node.elts))

    def visit_ListComp(self, node):
        raise NotImplementedError("List comprehensions are not supported")

    def visit_SetComp(self, node):
        raise NotImplementedError("Set comprehensions are not supported")

    def visit_DictComp(self, node):
        raise NotImplementedError("Dict comprehensions are not supported")

    def visit_GeneratorExp(self, node):
        self.active_ctx = ctx()
        return Visitor(
            self.source_file, inspect.Signature([]).bind(), self.globals, parent=self, function_name="<genexp>"
        ).run(node)

    def visit_Await(self, node):
        raise NotImplementedError("Await expressions are not supported")

    def visit_Yield(self, node):
        value = self.visit(node.value) if node.value else validate_value(None)
        ctx().scope.set_value("$yield", value)
        self.yield_ctxs.append(ctx())
        resume_ctx = ctx().new_disconnected()
        self.resume_ctxs.append(resume_ctx)
        set_ctx(resume_ctx)
        return validate_value(None)  # send() is unsupported, so yield returns None

    def visit_YieldFrom(self, node):
        value = self.visit(node.value)
        if isinstance(value, TupleImpl):
            for entry in value.value:
                ctx().scope.set_value("$yield", validate_value(entry))
                self.yield_ctxs.append(ctx())
                resume_ctx = ctx().new_disconnected()
                self.resume_ctxs.append(resume_ctx)
                set_ctx(resume_ctx)
            return validate_value(None)
        if not hasattr(value, "__iter__"):
            raise TypeError(f"Object of type '{type(value).__name__}' is not iterable")
        iterator = self.handle_call(node, value.__iter__)
        if not isinstance(iterator, SonolusIterator):
            raise ValueError("Expected a SonolusIterator")
        header = ctx().branch(None)
        set_ctx(header)
        result = self.handle_call(node, iterator.next)
        nothing_branch = ctx().branch(0)
        some_branch = ctx().branch(None)
        ctx().test = result._present.ir()
        set_ctx(some_branch)
        ctx().scope.set_value("$yield", result._value)
        self.yield_ctxs.append(ctx())
        resume_ctx = ctx().new_disconnected()
        self.resume_ctxs.append(resume_ctx)
        resume_ctx.outgoing[None] = header
        set_ctx(nothing_branch)
        return validate_value(None)

    def _has_real_method(self, obj: Value, method_name: str) -> bool:
        return hasattr(obj, method_name) and not isinstance(getattr(obj, method_name), MethodWrapperType)

    def visit_Compare(self, node):
        result_name = self.new_name("compare")
        ctx().scope.set_value(result_name, Num._accept_(0))
        l_val = self.visit(node.left)
        false_ctxs = []
        for i, (op, rhs) in enumerate(zip(node.ops, node.comparators, strict=True)):
            r_val = self.visit(rhs)
            inverted = isinstance(op, ast.NotIn)
            result = None
            if isinstance(op, ast.Is | ast.IsNot):
                if not (r_val._is_py_() and r_val._as_py_() is None):
                    raise TypeError("The right operand of 'is' must be None")
                if isinstance(op, ast.Is):
                    result = Num._accept_(l_val._is_py_() and l_val._as_py_() is None)
                else:
                    result = Num._accept_(not (l_val._is_py_() and l_val._as_py_() is None))
            elif type(op) in comp_ops and self._has_real_method(l_val, comp_ops[type(op)]):
                result = self.handle_call(node, getattr(l_val, comp_ops[type(op)]), r_val)
            if (
                (result is None or self.is_not_implemented(result))
                and type(op) in rcomp_ops
                and self._has_real_method(r_val, rcomp_ops[type(op)])
            ):
                result = self.handle_call(node, getattr(r_val, rcomp_ops[type(op)]), l_val)
            if result is None or self.is_not_implemented(result):
                # Can't defer to the default object.__eq__ or similar since reference equality (is) is not reliable
                if type(op) is ast.Eq and type(l_val) is not type(r_val):
                    return Num._accept_(False)
                elif type(op) is ast.NotEq and type(l_val) is not type(r_val):
                    return Num._accept_(True)
                else:
                    raise TypeError(
                        f"'{op_to_symbol[type(op)]}' not supported between instances of '{type(l_val).__name__}' and "
                        f"'{type(r_val).__name__}'"
                    )
            result = self.ensure_boolean_num(result)
            if inverted:
                result = result.not_()
            curr_ctx = ctx()
            if i == len(node.ops) - 1:
                curr_ctx.scope.set_value(result_name, result)
            elif result._is_py_():
                if result._as_py_():
                    l_val = r_val
                else:
                    false_ctxs.append(curr_ctx)
                    set_ctx(curr_ctx.into_dead())
                    break
            else:
                curr_ctx.test = result.ir()
                true_ctx = curr_ctx.branch(None)
                false_ctx = curr_ctx.branch(0)
                false_ctxs.append(false_ctx)
                set_ctx(true_ctx)
                l_val = r_val
        last_ctx = ctx()  # This is the result of the last comparison returning true
        set_ctx(Context.meet([last_ctx, *false_ctxs]))
        return ctx().scope.get_value(result_name)

    def visit_Call(self, node):
        from sonolus.script.internal.dict_impl import DictImpl

        fn = self.visit(node.func)
        args = []
        kwargs = {}
        for arg in node.args:
            if not ctx().live:
                return validate_value(None)
            if isinstance(arg, ast.Starred):
                args.extend(self.handle_starred(self.visit(arg.value)))
            else:
                args.append(self.visit(arg))
        for keyword in node.keywords:
            if not ctx().live:
                return validate_value(None)
            if keyword.arg:
                kwargs[keyword.arg] = self.visit(keyword.value)
            else:
                value = self.visit(keyword.value)
                if isinstance(value, DictImpl):
                    if not all(isinstance(k, str) for k in value.value):
                        raise ValueError("Keyword arguments must be strings")
                    kwargs.update(value.value)
                else:
                    raise ValueError("Starred keyword arguments (**kwargs) must be dictionaries")
        if not ctx().live:
            return validate_value(None)
        if fn._is_py_() and fn._as_py_() is _super and not args and not kwargs and "__class__" in self.globals:
            class_value = self.get_name("__class__")
            first_param_name = next(
                (
                    name
                    for name, param in self.bound_args.signature.parameters.items()
                    if param.kind in {inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD}
                ),
                None,
            )
            if first_param_name is not None:
                first_param_value = self.get_name(first_param_name)
                args = (validate_value(class_value), validate_value(first_param_value))
        return self.handle_call(node, fn, *args, **kwargs)

    def visit_FormattedValue(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_JoinedStr(self, node):
        raise NotImplementedError("F-strings are not supported")

    def visit_Constant(self, node):
        return validate_value(node.value)

    def visit_Attribute(self, node):
        return self.handle_getattr(node, self.visit(node.value), node.attr)

    def visit_Subscript(self, node):
        value = self.visit(node.value)
        slice_value = self.visit(node.slice)
        return self.handle_getitem(node, value, slice_value)

    def visit_Starred(self, node):
        raise NotImplementedError("Starred expressions are not supported")

    def visit_Name(self, node):
        return self.get_name(node.id)

    def get_name(self, name: str):
        if name in self.used_parent_binding_values:
            return self.used_parent_binding_values[name]
        self.active_ctx = ctx()
        v = self
        while v:
            if not isinstance(v.active_ctx.scope.get_binding(name), EmptyBinding):
                result = v.active_ctx.scope.get_value(name)
                if v is not self:
                    self.used_parent_binding_values[name] = result
                return result
            v = v.parent
        if name in self.globals:
            value = self.globals[name]
            if value is ctx:
                raise ValueError("Unexpected use of ctx in non meta-function")
            return validate_value(BUILTIN_IMPLS.get(id(value), value))
        raise NameError(f"Name {name} is not defined")

    def visit_List(self, node):
        raise NotImplementedError("List literals are not supported")

    def visit_Tuple(self, node):
        values = []
        for elt in node.elts:
            if isinstance(elt, ast.Starred):
                values.extend(self.handle_starred(self.visit(elt.value)))
            else:
                values.append(self.visit(elt))
        return validate_value(tuple(values))

    def visit_Slice(self, node):
        raise NotImplementedError("Slices are not supported")

    def handle_assign(self, target: ast.stmt | ast.expr, value: Value):
        match target:
            case ast.Name(id=name):
                ctx().scope.set_value(name, value)
            case ast.Attribute(value=attr_value, attr=attr):
                attr_value = self.visit(attr_value)
                self.handle_setattr(target, attr_value, attr, value)
            case ast.Subscript(value=sub_value, slice=slice_expr):
                sub_value = self.visit(sub_value)
                slice_value = self.visit(slice_expr)
                self.handle_setitem(target, sub_value, slice_value, value)
            case ast.Tuple(elts=elts) | ast.List(elts=elts):
                values = self.handle_starred(value)
                if len(elts) != len(values):
                    raise ValueError("Unpacking assignment requires the same number of elements")
                for elt, v in zip(elts, values, strict=False):
                    self.handle_assign(elt, validate_value(v))
            case ast.Starred():
                raise NotImplementedError("Starred assignment is not supported")
            case _:
                raise NotImplementedError("Unsupported assignment target")

    def handle_and(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)

        if l_val._is_py_():
            if l_val._as_py_():
                # The rhs is definitely evaluated, so we can return it directly
                return self.ensure_boolean_num(self.visit(r_expr))
            else:
                return l_val

        ctx_init.test = l_val.ir()
        res_name = self.new_name("and")

        set_ctx(ctx_init.branch(None))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        ctx().scope.set_value(res_name, Num._accept_(0))
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() and r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def handle_or(self, l_val: Value, r_expr: ast.expr) -> Value:
        ctx_init = ctx()
        l_val = self.ensure_boolean_num(l_val)

        if l_val._is_py_():
            if l_val._as_py_():
                return l_val
            else:
                # The rhs is definitely evaluated, so we can return it directly
                return self.ensure_boolean_num(self.visit(r_expr))

        ctx_init.test = l_val.ir()
        res_name = self.new_name("or")

        set_ctx(ctx_init.branch(None))
        ctx().scope.set_value(res_name, l_val)
        ctx_true = ctx()

        set_ctx(ctx_init.branch(0))
        r_val = self.ensure_boolean_num(self.visit(r_expr))
        ctx().scope.set_value(res_name, r_val)
        ctx_false = ctx()

        set_ctx(Context.meet([ctx_true, ctx_false]))
        if l_val._is_py_() and r_val._is_py_():
            return Num._accept_(l_val._as_py_() or r_val._as_py_())
        return ctx().scope.get_value(res_name)

    def generic_visit(self, node):
        if isinstance(node, ast.stmt | ast.expr):
            with self.reporting_errors_at_node(node):
                raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")
        raise NotImplementedError(f"Unsupported syntax: {type(node).__name__}")

    def handle_getattr(self, node: ast.stmt | ast.expr, target: Value, key: str) -> Value:
        # If this is changed, remember to update the getattr impl too
        with self.reporting_errors_at_node(node):
            if isinstance(target, ConstantValue):
                # Unwrap so we can access fields
                target = target._as_py_()
            descriptor = None
            for cls in type.mro(type(target)):
                descriptor = cls.__dict__.get(key, None)
                if descriptor is not None:
                    break
            match descriptor:
                case property(fget=getter):
                    return self.handle_call(node, getter, target)
                case SonolusDescriptor() | FunctionType() | classmethod() | staticmethod() | None:
                    return validate_value(getattr(target, key))
                case non_descriptor if not hasattr(non_descriptor, "__get__"):
                    return validate_value(getattr(target, key))
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_setattr(self, node: ast.stmt | ast.expr, target: Value, key: str, value: Value):
        # If this is changed, remember to update the setattr impl too
        with self.reporting_errors_at_node(node):
            if target._is_py_():
                target = target._as_py_()
            descriptor = None
            for cls in type.mro(type(target)):
                descriptor = cls.__dict__.get(key, None)
                if descriptor is not None:
                    break
            match descriptor:
                case property(fset=setter):
                    if setter is None:
                        raise AttributeError(f"Cannot set attribute {key} because property has no setter")
                    self.handle_call(node, setter, target, value)
                case SonolusDescriptor():
                    setattr(target, key, value)
                case _:
                    raise TypeError(f"Unsupported field or descriptor {key}")

    def handle_call[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R | Value:
        """Handles a call to the given callable."""
        self.active_ctx = ctx()
        debug_stack = self.active_ctx.callback_state.debug_stack
        try:
            debug_stack.append(f'File "{self.source_file}", line {node.lineno}, in {self.function_name}')
            if (
                isinstance(fn, Value)
                and fn._is_py_()
                and isinstance(fn._as_py_(), type)
                and issubclass(fn._as_py_(), Value)
            ):
                return validate_value(self.execute_at_node(node, fn._as_py_(), *args, **kwargs))
            else:
                return self.execute_at_node(node, lambda: validate_value(compile_and_call(fn, *args, **kwargs)))
        finally:
            debug_stack.pop()

    def handle_getitem(self, node: ast.stmt | ast.expr, target: Value, key: Value) -> Value:
        with self.reporting_errors_at_node(node):
            if target._is_py_() and isinstance(target._as_py_(), type):
                if not key._is_py_():
                    raise ValueError("Type parameters must be compile-time constants")
                return validate_value(target._as_py_()[key._as_py_()])
            else:
                if isinstance(target, Value) and hasattr(target, "__getitem__"):
                    return self.handle_call(node, target.__getitem__, key)
                raise TypeError(f"Cannot get items on {type(target).__name__}")

    def handle_setitem(self, node: ast.stmt | ast.expr, target: Value, key: Value, value: Value):
        with self.reporting_errors_at_node(node):
            if isinstance(target, Value) and hasattr(target, "__setitem__"):
                return self.handle_call(node, target.__setitem__, key, value)
            raise TypeError(f"Cannot set items on {type(target).__name__}")

    def handle_delitem(self, node: ast.stmt | ast.expr, target: Value, key: Value):
        with self.reporting_errors_at_node(node):
            if isinstance(target, Value) and hasattr(target, "__delitem__"):
                return self.handle_call(node, target.__delitem__, key)
            raise TypeError(f"Cannot delete items on {type(target).__name__}")

    def handle_starred(self, value: Value) -> tuple[Value, ...]:
        if isinstance(value, TupleImpl):
            return value.value
        raise ValueError("Unsupported starred expression")

    def is_not_implemented(self, value):
        value = validate_value(value)
        return value._is_py_() and value._as_py_() is NotImplemented

    def ensure_boolean_num(self, value) -> Num:
        # This just checks the type for now, although we could support custom __bool__ implementations in the future
        if not _is_num(value):
            raise TypeError(f"Invalid type where a bool (Num) was expected: {type(value).__name__}")
        return value

    def convert_to_boolean_num(self, node, value: Value) -> Num:
        if _is_num(value):
            return value
        if hasattr(type(value), "__bool__"):
            return self.ensure_boolean_num(self.handle_call(node, type(value).__bool__, validate_value(value)))
        if hasattr(type(value), "__len__"):
            length = self.handle_call(node, type(value).__len__, validate_value(value))
            if not _is_num(length):
                raise TypeError(f"Invalid type for __len__: {type(length).__name__}")
            if length._is_py_():
                return Num._accept_(length._as_py_() > 0)
            return length > Num._accept_(0)
        if isinstance(value, Record):
            return Num._accept_(1)
        # Not allowing other types to default to truthy for now in case there's any edge cases.
        raise TypeError(f"Converting {type(value).__name__} to bool is not supported")

    def arguments_to_signature(self, arguments: ast.arguments) -> inspect.Signature:
        parameters: list[inspect.Parameter] = []
        pos_only_count = len(arguments.posonlyargs)
        for i, arg in enumerate(arguments.posonlyargs):
            default_idx = i - pos_only_count + len(arguments.defaults)
            default = self.visit(arguments.defaults[default_idx]) if default_idx >= 0 else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.POSITIONAL_ONLY,
                default=default if default_idx >= 0 else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        pos_kw_count = len(arguments.args)
        for i, arg in enumerate(arguments.args):
            default_idx = i - pos_kw_count + len(arguments.defaults)
            default = self.visit(arguments.defaults[default_idx]) if default_idx >= 0 else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=default if default_idx >= 0 else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        if arguments.vararg:
            param = inspect.Parameter(
                name=arguments.vararg.arg,
                kind=inspect.Parameter.VAR_POSITIONAL,
                default=inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        for i, arg in enumerate(arguments.kwonlyargs):
            default = self.visit(arguments.kw_defaults[i]) if arguments.kw_defaults[i] is not None else None
            param = inspect.Parameter(
                name=arg.arg,
                kind=inspect.Parameter.KEYWORD_ONLY,
                default=default if default is not None else inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        if arguments.kwarg:
            param = inspect.Parameter(
                name=arguments.kwarg.arg,
                kind=inspect.Parameter.VAR_KEYWORD,
                default=inspect.Parameter.empty,
                annotation=inspect.Parameter.empty,
            )
            parameters.append(param)

        return inspect.Signature(parameters)

    def raise_exception_at_node(self, node: ast.stmt | ast.expr, cause: Exception) -> Never:
        """Throws a compilation error at the given node."""

        def thrower() -> Never:
            raise CompilationError(str(cause)) from cause

        self.execute_at_node(node, thrower)

    def execute_at_node[**P, R](
        self, node: ast.stmt | ast.expr, fn: Callable[P, R], /, *args: P.args, **kwargs: P.kwargs
    ) -> R:
        """Executes the given function at the given node for a better traceback."""
        if ctx().no_eval:
            return fn(*args, **kwargs)

        location_args = {
            "lineno": node.lineno,
            "col_offset": node.col_offset,
            "end_lineno": node.end_lineno,
            "end_col_offset": node.end_col_offset,
        }

        expr = ast.Expression(
            body=ast.Call(
                func=ast.Name(id="fn", ctx=ast.Load(), **location_args),
                args=[
                    ast.Starred(
                        value=ast.Name(id="args", ctx=ast.Load(), **location_args),
                        ctx=ast.Load(),
                        **location_args,
                    )
                ],
                keywords=[
                    ast.keyword(
                        value=ast.Name(id="kwargs", ctx=ast.Load(), **location_args),
                        arg=None,
                        **location_args,
                    )
                ],
                **location_args,
            )
        )
        return eval(
            compile(expr, filename=self.source_file, mode="eval").replace(co_name=self.function_name),
            {"fn": fn, "args": args, "kwargs": kwargs, "_filter_traceback_": True},
        )

    def reporting_errors_at_node(self, node: ast.stmt | ast.expr):
        return ReportingErrorsAtNode(self, node)

    def new_name(self, name: str):
        self.used_names[name] = self.used_names.get(name, 0) + 1
        return f"${name}_{self.used_names[name]}"


# Not using @contextmanager so it doesn't end up in tracebacks
class ReportingErrorsAtNode:
    def __init__(self, compiler, node: ast.stmt | ast.expr):
        self.compiler = compiler
        self.node = node

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            return

        if issubclass(exc_type, CompilationError):
            raise exc_value from exc_value.__cause__

        if exc_value is not None:
            self.compiler.raise_exception_at_node(self.node, exc_value)


class Generator(TransientValue, SonolusIterator):
    def __init__(
        self,
        return_test: Num,
        entry: Context,
        exit_: Context,
        value: Maybe,
        used_bindings: dict[str, Value],
        parent: Visitor,
    ):
        self.i = 0
        self.return_test = return_test
        self.entry = entry
        self.exit = exit_
        self.value = value
        self.used_bindings = used_bindings
        self.parent = parent

    @meta_fn
    def next(self):
        self._validate_bindings()
        self.return_test._set_(self.i)
        after_ctx = ctx().new_disconnected()
        ctx().outgoing[None] = self.entry
        self.exit.outgoing[self.i] = after_ctx
        self.i += 1
        set_ctx(after_ctx)
        return self.value

    def _validate_bindings(self):
        for key, value in self.used_bindings.items():
            v = self.parent
            while v:
                if not isinstance(v.active_ctx.scope.get_binding(key), EmptyBinding):
                    result = v.active_ctx.scope.get_value(key)
                    if result is not value:
                        raise ValueError(f"Binding '{key}' has been modified since the generator was created")
                v = v.parent

    def __iter__(self):
        return self

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls)

    @classmethod
    def _accept_(cls, value: Any) -> Generator:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value of type {type(value).__name__} as {cls.__name__}")
        return value

    def _is_py_(self) -> bool:
        return False

    def _as_py_(self) -> Any:
        raise NotImplementedError



import itertools
from math import pi

from hypothesis import assume, given
from hypothesis import strategies as st

from sonolus.script.quad import Quad, Rect
from sonolus.script.vec import Vec2
from tests.script.conftest import is_close, run_and_validate

floats = st.floats(min_value=-9, max_value=9, allow_nan=False, allow_infinity=False)
nonzero_floats = floats.filter(lambda x: abs(x) > 1e-2)


@st.composite
def vecs(draw):
    x = draw(floats)
    y = draw(floats)
    return Vec2(x, y)


def _area(points: list[Vec2]) -> float:
    return 0.5 * abs(sum(p.x * q.y - p.y * q.x for p, q in zip(points, points[1:] + points[:1], strict=True)))


@st.composite
def quads(draw):
    points = [draw(vecs()) for _ in range(4)]
    for p1, p2 in itertools.combinations(points, 2):
        assume((p1 - p2).magnitude > 1e-2)
    centroid = sum(points, Vec2(0, 0)) / 4
    points = sorted(points, key=lambda p: (p - centroid).angle)
    assume(_area(points) > 1e-2)
    return Quad(*points)


@st.composite
def rects(draw):
    l = draw(floats)
    r = draw(floats)
    if l > r:
        l, r = r, l

    b = draw(floats)
    t = draw(floats)
    if b > t:
        b, t = t, b

    return Rect(t=t, r=r, b=b, l=l)


@given(
    quad=quads(),
    translation=vecs(),
)
def test_quad_translate(quad, translation):
    def fn():
        return quad.translate(translation)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl + translation)
    assert is_close(result.tl, quad.tl + translation)
    assert is_close(result.tr, quad.tr + translation)
    assert is_close(result.br, quad.br + translation)


@given(
    quad=quads(),
    factor=vecs(),
)
def test_quad_scale(quad, factor):
    def fn():
        return quad.scale(factor)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl * factor)
    assert is_close(result.tl, quad.tl * factor)
    assert is_close(result.tr, quad.tr * factor)
    assert is_close(result.br, quad.br * factor)


@given(
    quad=quads(),
    factor=vecs(),
    pivot=vecs(),
)
def test_quad_scale_about(quad, factor, pivot):
    def fn():
        return quad.scale_about(factor, pivot)

    result = run_and_validate(fn)
    assert is_close(result.bl, (quad.bl - pivot) * factor + pivot)
    assert is_close(result.tl, (quad.tl - pivot) * factor + pivot)
    assert is_close(result.tr, (quad.tr - pivot) * factor + pivot)
    assert is_close(result.br, (quad.br - pivot) * factor + pivot)


@given(
    quad=quads(),
    angle=floats,
)
def test_quad_rotate(quad, angle):
    def fn():
        return quad.rotate(angle)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl.rotate(angle))
    assert is_close(result.tl, quad.tl.rotate(angle))
    assert is_close(result.tr, quad.tr.rotate(angle))
    assert is_close(result.br, quad.br.rotate(angle))


@given(
    quad=quads(),
    angle=floats,
    pivot=vecs(),
)
def test_quad_rotate_about(quad, angle, pivot):
    def fn():
        return quad.rotate_about(angle, pivot)

    result = run_and_validate(fn)
    assert is_close(result.bl, quad.bl.rotate_about(angle, pivot))
    assert is_close(result.tl, quad.tl.rotate_about(angle, pivot))
    assert is_close(result.tr, quad.tr.rotate_about(angle, pivot))
    assert is_close(result.br, quad.br.rotate_about(angle, pivot))


@given(quad=quads())
def test_quad_center(quad):
    def fn():
        return quad.center

    result = run_and_validate(fn)
    expected = (quad.bl + quad.tr + quad.tl + quad.br) / 4
    assert is_close(result, expected)


def test_quad_permute():
    for rotation in range(-5, 6):
        quad = Quad(
            Vec2(-1, -1),
            Vec2(-1, 1),
            Vec2(1, 1),
            Vec2(1, -1),
        )

        def fn():
            return quad.permute(rotation)  # noqa: B023

        result = run_and_validate(fn)
        expected = quad.rotate(rotation * pi / 2)
        assert is_close(result.bl, expected.bl)
        assert is_close(result.tl, expected.tl)
        assert is_close(result.tr, expected.tr)
        assert is_close(result.br, expected.br)


# Rect Tests
@given(
    rect=rects(),
    translation=vecs(),
)
def test_rect_translate(rect, translation):
    def fn():
        return rect.translate(translation)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t + translation.y)
    assert is_close(result.r, rect.r + translation.x)
    assert is_close(result.b, rect.b + translation.y)
    assert is_close(result.l, rect.l + translation.x)


@given(
    rect=rects(),
    factor=vecs(),
)
def test_rect_scale(rect, factor):
    def fn():
        return rect.scale(factor)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t * factor.y)
    assert is_close(result.r, rect.r * factor.x)
    assert is_close(result.b, rect.b * factor.y)
    assert is_close(result.l, rect.l * factor.x)


@given(
    rect=rects(),
    factor=vecs(),
    pivot=vecs(),
)
def test_rect_scale_about(rect, factor, pivot):
    def fn():
        return rect.scale_about(factor, pivot)

    result = run_and_validate(fn)
    assert is_close(result.t, (rect.t - pivot.y) * factor.y + pivot.y)
    assert is_close(result.r, (rect.r - pivot.x) * factor.x + pivot.x)
    assert is_close(result.b, (rect.b - pivot.y) * factor.y + pivot.y)
    assert is_close(result.l, (rect.l - pivot.x) * factor.x + pivot.x)


@given(
    rect=rects(),
    expansion=vecs(),
)
def test_rect_expand(rect, expansion):
    def fn():
        return rect.expand(expansion)

    result = run_and_validate(fn)
    assert is_close(result.t, rect.t + expansion.y)
    assert is_close(result.r, rect.r + expansion.x)
    assert is_close(result.b, rect.b - expansion.y)
    assert is_close(result.l, rect.l - expansion.x)


@given(
    rect=rects(),
    point=vecs(),
)
def test_rect_contains_point(rect, point):
    def fn():
        return rect.contains_point(point)

    result = run_and_validate(fn)
    expected = rect.l <= point.x <= rect.r and rect.b <= point.y <= rect.t
    assert is_close(result, expected)


@given(
    center=vecs(),
    dimensions=vecs(),
)
def test_rect_from_center(center, dimensions):
    def fn():
        return Rect.from_center(center, dimensions)

    result = run_and_validate(fn)
    assert is_close(result.center.x, center.x)
    assert is_close(result.center.y, center.y)
    assert is_close(result.w, dimensions.x)
    assert is_close(result.h, dimensions.y)


@given(rect=rects())
def test_rect_as_quad(rect):
    def fn():
        return rect.as_quad()

    result = run_and_validate(fn)
    assert is_close(result.bl, rect.bl)
    assert is_close(result.tl, rect.tl)
    assert is_close(result.tr, rect.tr)
    assert is_close(result.br, rect.br)


@st.composite
def quad_and_point(draw):
    quad = draw(quads())
    min_x = min(quad.bl.x, quad.tl.x, quad.tr.x, quad.br.x)
    max_x = max(quad.bl.x, quad.tl.x, quad.tr.x, quad.br.x)
    min_y = min(quad.bl.y, quad.tl.y, quad.tr.y, quad.br.y)
    max_y = max(quad.bl.y, quad.tl.y, quad.tr.y, quad.br.y)
    x = draw(st.floats(min_value=min_x, max_value=max_x))
    y = draw(st.floats(min_value=min_y, max_value=max_y))
    point = Vec2(x, y)
    return quad, point


@given(quad_point=quad_and_point())
def test_quad_contains_point(quad_point):
    quad, point = quad_point

    def fn():
        return quad.contains_point(point)

    # Don't have an easy way to validate the correct solution
    # But this still checks that the compiled and Python versions are consistent
    run_and_validate(fn)


def test_rect_from_margin_one_arg():
    def fn():
        return Rect.from_margin(1)

    result = run_and_validate(fn)
    assert result.t == 1
    assert result.r == 1
    assert result.b == -1
    assert result.l == -1
    assert result.w == 2
    assert result.h == 2


def test_rect_from_margin_two_args():
    def fn():
        return Rect.from_margin(1, 2)

    result = run_and_validate(fn)
    assert result.t == 1
    assert result.r == 2
    assert result.b == -1
    assert result.l == -2
    assert result.w == 4
    assert result.h == 2


def test_rect_from_margin_three_args():
    def fn():
        return Rect.from_margin(1, 2, 3)

    result = run_and_validate(fn)
    assert result.t == 1
    assert result.r == 2
    assert result.b == -3
    assert result.l == -2
    assert result.w == 4
    assert result.h == 4


def test_rect_from_margin_four_args():
    def fn():
        return Rect.from_margin(1, 2, 3, 4)

    result = run_and_validate(fn)
    assert result.t == 1
    assert result.r == 2
    assert result.b == -3
    assert result.l == -4
    assert result.w == 6
    assert result.h == 4


def test_quad_mt():
    quad = Quad(
        Vec2(0, 0),  # bl
        Vec2(0, 2),  # tl
        Vec2(2, 2),  # tr
        Vec2(2, 0),  # br
    )

    def fn():
        return quad.mt

    result = run_and_validate(fn)
    assert result.x == 1
    assert result.y == 2


def test_quad_mr():
    quad = Quad(
        Vec2(0, 0),  # bl
        Vec2(0, 2),  # tl
        Vec2(2, 2),  # tr
        Vec2(2, 0),  # br
    )

    def fn():
        return quad.mr

    result = run_and_validate(fn)
    assert result.x == 2
    assert result.y == 1


def test_quad_mb():
    quad = Quad(
        Vec2(0, 0),  # bl
        Vec2(0, 2),  # tl
        Vec2(2, 2),  # tr
        Vec2(2, 0),  # br
    )

    def fn():
        return quad.mb

    result = run_and_validate(fn)
    assert result.x == 1
    assert result.y == 0


def test_quad_ml():
    quad = Quad(
        Vec2(0, 0),  # bl
        Vec2(0, 2),  # tl
        Vec2(2, 2),  # tr
        Vec2(2, 0),  # br
    )

    def fn():
        return quad.ml

    result = run_and_validate(fn)
    assert result.x == 0
    assert result.y == 1


def test_rect_mt():
    rect = Rect(t=3, r=4, b=1, l=2)

    def fn():
        return rect.mt

    result = run_and_validate(fn)
    assert result.x == 3
    assert result.y == 3


def test_rect_mr():
    rect = Rect(t=3, r=4, b=1, l=2)

    def fn():
        return rect.mr

    result = run_and_validate(fn)
    assert result.x == 4
    assert result.y == 2


def test_rect_mb():
    rect = Rect(t=3, r=4, b=1, l=2)

    def fn():
        return rect.mb

    result = run_and_validate(fn)
    assert result.x == 3
    assert result.y == 1


def test_rect_ml():
    rect = Rect(t=3, r=4, b=1, l=2)

    def fn():
        return rect.ml

    result = run_and_validate(fn)
    assert result.x == 2
    assert result.y == 2



import textwrap
from typing import NamedTuple

from sonolus.backend.ops import Op

type EngineNode = int | float | FunctionNode


class FunctionNode(NamedTuple):
    func: Op
    args: tuple[EngineNode, ...]


def format_engine_node(node: EngineNode) -> str:
    if isinstance(node, FunctionNode):
        match len(node.args):
            case 0:
                return f"{node.func.name}()"
            case 1:
                return f"{node.func.name}({format_engine_node(node.args[0])})"
            case _:
                return f"{node.func.name}(\n{
                    textwrap.indent('\n'.join(format_engine_node(arg) for arg in node.args), '  ')
                }\n)"
    else:
        return str(node)



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import ArraySet
from sonolus.script.debug import assert_true
from tests.script.conftest import run_and_validate

ints = st.integers(min_value=-999, max_value=999)
sets = st.sets(ints, min_size=1, max_size=20)


@st.composite
def set_and_present_value(draw):
    values = draw(sets)
    value = draw(st.sampled_from(list(values)))
    return values, value


@st.composite
def set_and_missing_value(draw):
    values = draw(sets)
    missing = draw(ints.filter(lambda x: x not in values))
    return values, missing


def test_array_set_add_basic():
    def fn():
        s = ArraySet[int, 4].new()
        assert_true(s.add(2))
        assert_true(s.add(4))
        assert_true(2 in s)
        assert_true(4 in s)
        assert_true(6 not in s)
        return s

    assert sorted(run_and_validate(fn)) == [2, 4]


def test_array_set_clear():
    def fn():
        s = ArraySet[int, 4].new()
        s.add(2)
        s.add(4)
        s.add(6)
        s.clear()
        return s._values

    assert list(run_and_validate(fn)) == []


@given(set_and_present_value())
def test_array_set_add_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        s = ArraySet[int, value_count + 1].new()
        for v in values:
            s.add(v)
        assert_true(not s.add(value))
        return s

    assert sorted(run_and_validate(fn)) == sorted(values)


@given(set_and_missing_value())
def test_array_set_add_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        s = ArraySet[int, value_count + 1].new()
        for v in values:
            s.add(v)
        assert_true(s.add(missing))
        return s

    assert sorted(run_and_validate(fn)) == sorted([*list(values), missing])


@given(set_and_present_value())
def test_array_set_remove_present(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        s = ArraySet[int, value_count].new()
        for v in values:
            s.add(v)
        assert_true(s.remove(value))
        return s

    assert sorted(run_and_validate(fn)) == sorted(value_set - {value})


@given(set_and_missing_value())
def test_array_set_remove_missing(args):
    value_set, missing = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        s = ArraySet[int, value_count].new()
        for v in values:
            s.add(v)
        assert_true(not s.remove(missing))
        return s

    assert sorted(run_and_validate(fn)) == sorted(value_set)


@given(set_and_missing_value())
def test_array_set_add_remove_round_trip(args):
    value_set, value = args
    values = Array(*value_set)
    value_count = len(value_set)

    def fn():
        s = ArraySet[int, value_count + 1].new()
        for v in values:
            s.add(v)
        assert_true(s.add(value))
        assert_true(s.remove(value))
        return s

    assert sorted(run_and_validate(fn)) == sorted(value_set)


def test_array_set_full():
    def fn():
        s = ArraySet[int, 4].new()
        assert_true(s.add(2))
        assert_true(s.add(4))
        assert_true(s.add(6))
        assert_true(s.add(8))
        assert_true(not s.add(10))  # Set is full
        return s

    assert list(run_and_validate(fn)) == [2, 4, 6, 8]


def test_array_set_truthiness_empty():
    def fn():
        x = ArraySet[int, 5].new()
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_array_set_truthiness_non_empty():
    def fn():
        x = ArraySet[int, 5].new()
        x.add(42)
        return 1 if x else 0

    assert run_and_validate(fn) == 1



from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from tests.script.conftest import run_and_validate

ints = st.integers(min_value=-1000, max_value=1000)
floats = st.floats(min_value=-99999, max_value=99999, allow_nan=False, allow_infinity=False)


def test_num_basic():
    def fn():
        return 1

    assert run_and_validate(fn) == 1


@given(n=floats)
def test_num_unary(n):
    def fn():
        plus = +n
        minus = -n
        return Array(plus, minus)

    assert run_and_validate(fn) == Array(+n, -n)


@given(
    a=floats,
    b=floats,
)
def test_num_comparison(a, b):
    def fn():
        lt = a < b
        le = a <= b
        eq = a == b
        ne = a != b
        ge = a >= b
        gt = a > b
        return Array(lt, le, eq, ne, ge, gt)

    assert run_and_validate(fn) == Array(a < b, a <= b, a == b, a != b, a >= b, a > b)


def are_valid_pow_operands(a, b):
    if b > 10:
        # Avoid huge numbers
        return False
    if a < 0 and b % 1 != 0:
        # Avoid complex numbers
        return False
    if a == 0 and b <= 0:
        # Avoid division by zero
        return False
    if abs(a) < 1e-6 and b <= 0:
        # Avoid division by zero or by very small numbers
        return False
    if a < 1 and b < -10:  # noqa: SIM103
        # Avoid division by very small numbers
        return False
    return True


@given(
    a=floats,
    b=floats,
)
def test_num_binary(a, b):
    def fn():
        add = a + b
        sub = a - b
        mul = a * b
        div = a / b if abs(b) > 1e-6 else 0
        mod = a % b if abs(b) > 1e-6 else 0
        power = a**b if are_valid_pow_operands(a, b) else 0
        return Array(add, sub, mul, div, mod, power)

    assert run_and_validate(fn) == Array(
        a + b,
        a - b,
        a * b,
        a / b if abs(b) > 1e-6 else 0,
        a % b if abs(b) > 1e-6 else 0,
        a**b if are_valid_pow_operands(a, b) else 0,
    )


@given(
    a=ints,
    b=ints.filter(lambda x: x != 0),
)
def test_num_floordiv(a, b):
    # floordiv can behave weirdly with float operands due to precision issues, so
    # floor(a / b) may not equal a // b.
    def fn():
        return a // b

    assert run_and_validate(fn) == a // b


@given(
    a=floats,
    b=floats,
)
def test_num_augmented(a, b):
    def fn():
        add = a
        add += b
        sub = a
        sub -= b
        mul = a
        mul *= b
        div = a
        if b != 0:
            div /= b
        mod = a
        if b != 0:
            mod %= b
        power = a
        if are_valid_pow_operands(a, b):
            power **= b
        floordiv = a
        if b != 0:
            floordiv //= b
        return Array(add, sub, mul, div, mod, power, floordiv)

    assert run_and_validate(fn) == Array(
        a + b,
        a - b,
        a * b,
        a / b if b != 0 else a,
        a % b if b != 0 else a,
        a**b if are_valid_pow_operands(a, b) else a,
        a // b if b != 0 else a,
    )


# What we care about is that compiled behavior matches vanilla Python behavior
# So run_and_validate is more or less enough.


@given(n=ints)
def test_num_while_assignment(n):
    def fn():
        result = 0
        i = 0
        while i < n:
            result += i
            i += 1
        return result

    assert run_and_validate(fn) == sum(range(n))


@given(n=ints)
def test_num_while_else(n):
    def fn():
        result = 0
        i = 0
        while i < n:
            result += i
            i += 1
        result += 100
        return result

    assert run_and_validate(fn) == sum(range(n)) + 100


@given(
    a=ints,
    b=ints,
)
def test_num_while_break(a, b):
    def fn():
        result = 0
        i = 0
        while i < a:
            result += i
            i += 1
            if i == b:
                break
        else:
            result = -1
        return result

    run_and_validate(fn)


@given(
    a=ints,
    b=ints,
)
def test_num_while_continue(a, b):
    def fn():
        result = 0
        i = 0
        while i < a:
            i += 1
            if i == b:
                result += 100
                continue
            result += i - 1
        return result

    run_and_validate(fn)
